import bpy

def mn():
    """Return the MolecularNodes Collection
    
    The collection called 'MolecularNodes' inside the Blender scene is returned. If the 
    collection does not exist first, it is created.
    """
    coll = bpy.data.collections.get('MolecularNodes')
    if not coll:
        coll = bpy.data.collections.new('MolecularNodes')
        bpy.context.scene.collection.children.link(coll)
    return coll

def armature(name = 'MN_armature'):
    coll = bpy.data.collections.get(name)
    if not coll:
        coll = bpy.data.collections.new(name)
        mn().children.link(coll)
    return coll

def data(suffix = ""):
    """A collection for storing MN related data objects.
    """
    name = f"MN_data{suffix}"
    
    collection = bpy.data.collections.get(name)
    if not collection:
        collection = bpy.data.collections.new(name)
        mn().children.link(collection)
        
        # disable the view of the data collection
        bpy.context.view_layer.layer_collection.children['MolecularNodes'].children[name].exclude = True
    return collection

def frames(name="", parent=None, suffix="_frames"):
    """Create a Collection for Frames of a Trajectory

    Args:
        name (str, optional): Name of the collection for the frames. Defaults to "".
        parent (_type_, optional): A blender collection which will become the parent 
        collection. Defaults to the MolecularNodes collection if None.
    """
    coll_frames = bpy.data.collections.new(name + suffix)
    if not parent:
        mn().children.link(coll_frames)
    else:
        parent.children.link(coll_frames)
    
    return coll_frames

def cellpack(name="", parent=None, fallback=False):
    full_name = f"cellpack_{name}"
    coll = bpy.data.collections.get(full_name)
    if coll and fallback:
        return coll
    
    coll = bpy.data.collections.new(full_name)
    
    if parent:
        parent.children.link(coll)
    else:
        data().children.link(coll)
    
    return coll



import bpy
import numpy as np
from . import obj, coll

def clear_armature(object):
    for mod in object.modifiers:
        if mod.type == "ARMATURE":
            if mod.object:
                bpy.data.objects.remove(mod.object)
            object.modifiers.remove(mod)

def add_bones(object, name = 'armature'):
    ## creates bones and assigns correct weights
    
    clear_armature(object)
    
    bone_positions, bone_weights, chain_ids = get_bone_positions(object)
    
    armature = create_bones(bone_positions, chain_ids)
    for i in range(bone_weights.shape[1]):
        group = object.vertex_groups.new(name=f'mn_armature_{i}')
        vertex_indices = np.where(bone_weights[:, i] == 1)[0]
        group.add(vertex_indices.tolist(), 1, 'ADD')
    
    object.select_set(True)
    armature.select_set(True)
    bpy.context.view_layer.objects.active = armature
    bpy.ops.object.parent_set(type='ARMATURE')
    
    bpy.context.view_layer.objects.active = object
    bpy.ops.object.modifier_move_to_index('EXEC_DEFAULT', modifier="Armature", index=0)

    return armature

def get_bone_positions(object):
    
    positions, atom_name, chain_id, res_id, sec_struct = [
        obj.get_attribute(object, att) 
        for att in ['position', 'atom_name', 'chain_id', 'res_id', 'sec_struct']
        ]
    
    is_alpha_carbon = atom_name == 2
    idx = np.where(is_alpha_carbon)[0]
    bone_positions = positions[idx, :]
    bone_positions = np.vstack((bone_positions, positions[-1]))
    group_ids = np.cumsum(is_alpha_carbon)
    groups = np.unique(group_ids)
    bone_weights = np.zeros((len(group_ids), len(groups)))

    for i, unique_id in enumerate(groups):
        bone_weights[:, i] = ((group_ids - 1) == unique_id).astype(int)

    print("get_bone_positions")
    return bone_positions, bone_weights, chain_id[idx]

def get_bone_weights(object):
    print('hello world')

def create_bones(positions, chain_ids, name = 'armature'):
    
    bpy.ops.object.add(type='ARMATURE', enter_editmode=True)
    object = bpy.context.active_object
    object.name = name
    coll.armature().objects.link(object)
    armature = object.data
    armature.name = f'{name}_frame'
    arm_name = armature.name
    bones = []
    # add bones
    for i, position in enumerate(positions):
        try:
            pos_a = position
            pos_b = positions[i + 1, :]
        except:
            continue

        bone_name = f"mn_armature_{i}"
        bone = armature.edit_bones.new(bone_name)
        bone.head = pos_a
        bone.tail = pos_b
        bones.append(bone.name)

    armature = bpy.data.armatures[arm_name]
    bones_a = bones.copy()
    bones_b = bones.copy()
    bones_b.pop(0)
    bones = zip(bones_a, bones_b)

    for bone_a, bone_b in bones:
        armature.edit_bones.active = armature.edit_bones[bone_a]
        for bone in [bone_a, bone_b]:
            armature.edit_bones[bone].select = True
        bpy.ops.armature.parent_set(type='CONNECTED')
        for bone in [bone_a, bone_b]:
            armature.edit_bones[bone].select = False
    bpy.ops.object.editmode_toggle()
    
    return object

class MN_MT_Add_Armature(bpy.types.Operator):
    bl_idname = 'mn.add_armature'
    bl_label = 'Add Armature'
    bl_description = 'Automatically add armature for each amino acid of the structure   '
    
    def execute(self, context):
        object = context.active_object
        add_bones(bpy.data.objects[object.name], name = object.name)
        
        return {'FINISHED'}


import bpy
from .. import pkg
from ..ui import pref
from ..blender import nodes
from ..io import (
    pdb, local, star, cellpack, md, density, dna
)

bpy.types.Scene.MN_panel = bpy.props.EnumProperty(
    name = "Panel Selection", 
    items = (
        ('import', "Import", "Import macromolecules", 0),
        ('object', "Object", "Adjust settings affecting the selected object.", 1),
        ('scene',  "Scene", "Change settings for the world and rendering", 2)
    )
)

bpy.types.Scene.MN_panel_import = bpy.props.EnumProperty(
    name = "Method", 
    items = (
        ('pdb', "PDB", "Download from the PDB", 0),
        ('local', "Local", "Open a local file", 1),
        ('md', "MD", "Import a molecular dynamics trajectory", 2),
        ('density', "Density", "Import an EM Density Map", 3), 
        ('star', 'Starfile', "Import a .starfile mapback file", 4), 
        ('cellpack', 'CellPack', "Import a CellPack .cif/.bcif file.", 5), 
        ('dna', 'oxDNA', 'Import an oxDNA fil.', 6)
    )
)

chosen_panel = {
    'pdb': pdb, 
    'local': local, 
    'star': star, 
    'md': md, 
    'density': density, 
    'cellpack': cellpack,
    'dna': dna
    
}

packages = {
    'pdb': ['biotite', 'scipy'], 
    'star': ['starfile', 'eulerangles'], 
    'local': ['biotite', 'scipy'], 
    'cellpack': ['biotite', 'msgpack', 'scipy'], 
    'md': ['MDAnalysis'], 
    'density': ['mrcfile', 'scipy'],
    'dna': []
}

class MN_OT_Change_Style(bpy.types.Operator):
    bl_idname = 'mn.style_change'
    bl_label = 'Style'
    
    style: bpy.props.EnumProperty(
        name = "Style", 
        items = nodes.STYLE_ITEMS
    )
    
    def execute(self, context):
        object = context.active_object
        nodes.change_style_node(object, self.style)
        
        return {'FINISHED'}

def check_installs(selection):
    for package in packages[selection]:
        if not pkg.is_current(package):
            return False
    
    return True

def panel_import(layout, context):
    scene = context.scene
    selection = scene.MN_panel_import
    layout.prop(scene, 'MN_panel_import')
    
    apple_warning = pkg._is_apple_silicon and selection == "md" and not pkg.is_current('MDAnalysis')
    install_required = not check_installs(selection) or apple_warning
    
    if apple_warning:
        pref.apple_silicon_warning(layout)
    
    buttons = layout.column(align=True)
    buttons.enabled = not apple_warning
    
    if install_required:
        buttons.label(text = 'Please install the requried packages.')
        for package in packages[selection]:
            pkg.button_install_pkg(buttons, package, pkg.get_pkgs()[package]['version'])
    
    col = layout.column()
    col.enabled = not install_required
    chosen_panel[selection].panel(col, scene)


def ui_from_node(layout, node):
    """
    Generate the UI for a particular node, which displays the relevant node inputs
    for user control in a panel, rather than through the node editor.
    """
    col = layout.column(align = True)
    ntree = bpy.context.active_object.modifiers['MolecularNodes'].node_group
    
    tree = node.node_tree.interface.items_tree
    
    for item in tree.values():
        if item.item_type == "PANEL":
            col.label(text=item.name)
        elif item.name == "Selection":
            continue
        else:
            if item.in_out != "INPUT":
                continue
            if item.socket_type == "NodeSocketGeometry":
                continue
            # col.prop(node.inputs[item.identifier], 'default_value', text = item.name)
            col.template_node_view(ntree, node, node.inputs[item.identifier])

def panel_object(layout, context):
    scene = context.scene
    object = context.active_object
    mol_type = object.mn.molecule_type
    if mol_type == "":
        layout.label(text = "No MN object selected.")
        return None
    
    if mol_type == "pdb":
        layout.label(text = f"PDB: {object.mn.pdb_code.upper()}")
    if mol_type == "md":
        layout.prop(object.mn, 'subframes')
        
    
    row = layout.row(align=True)
    row.label(text = "Style")
    current_style = nodes.format_node_name(nodes.get_style_node(object).node_tree.name).replace("Style ", "")
    row.operator_menu_enum('mn.style_change', 'style', text = current_style)
    box = layout.box()
    ui_from_node(box, nodes.get_style_node(object))
    # layout.label(text='Color')
    # box = layout.box()
    # ui_from_node(box, nodes.get_color_node(object))
    row = layout.row()
    row.label(text="Experimental", icon_value=2)
    row.operator('mn.add_armature')


def panel_scene(layout, context):
    scene = context.scene
    
    cam = bpy.data.cameras[bpy.data.scenes["Scene"].camera.name]
    world_shader = bpy.data.worlds["World Shader"].node_tree.nodes["MN_world_shader"]
    grid = layout.grid_flow()
    col = grid.column()
    col.label(text = "World Settings")
    world = col.box()
    world.prop(bpy.data.scenes["Scene"].render, "engine")
    if scene.render.engine == "CYCLES":
        world.prop(bpy.data.scenes["Scene"].cycles, "samples")
    else:
        world.prop(bpy.data.scenes["Scene"].eevee, "taa_render_samples")
    world.label(text = "Background")
    world.prop(world_shader.inputs[1], 'default_value', text = "World Lighting")
    row = world.row()
    row.prop(scene.render, 'film_transparent')
    row.prop(world_shader.inputs[2], 'default_value', text = "")
    
    col = grid.column()
    col.label(text="Camera Settings")
    camera = col.box()
    camera.prop(cam, "lens")
    col = camera.column(align=True)
    row = col.row(align=True)
    row.prop(bpy.data.scenes["Scene"].render, "resolution_x", text = "X")
    row.prop(bpy.data.scenes["Scene"].render, "resolution_y", text = "Y")
    row = camera.grid_flow()
    row.prop(cam.dof, 'use_dof')
    row.prop(bpy.data.scenes["Scene"].render, "use_motion_blur")
    focus = camera.column()
    focus.enabled = cam.dof.use_dof
    focus.prop(cam.dof, 'focus_object')
    distance = focus.row()
    distance.enabled = (cam.dof.focus_object is None)
    distance.prop(cam.dof, 'focus_distance')
    focus.prop(cam.dof, 'aperture_fstop')


class MN_PT_panel(bpy.types.Panel):
    bl_label = 'Molecular Nodes'
    bl_idname = 'MN_PT_panel'
    bl_space_type = 'PROPERTIES'
    bl_region_type = 'WINDOW'
    bl_context = 'scene'
    bl_order = 0
    bl_options = {'HEADER_LAYOUT_EXPAND'}
    bl_ui_units_x=0

    def draw(self, context):
        layout = self.layout
        scene = context.scene
        row = layout.row(align=True)
        for p in ['import', 'object', 'scene']:
            row.prop_enum(scene, 'MN_panel', p)
        
        # the possible panel functions to choose between
        which_panel = {
            "import": panel_import,
            "scene": panel_scene,
            "object": panel_object
        }
        # call the required panel function with the layout and context
        which_panel[scene.MN_panel](layout, context)



import bpy

from .node_info import menu_items
from .func import build_menu


class MN_MT_Node_Color(bpy.types.Menu):
    bl_idname = 'MN_MT_NODE_COLOR'
    bl_label = ''
    
    def draw(self, context):
        layout = self.layout
        build_menu(layout, menu_items['color'])


class MN_MT_Node_Bonds(bpy.types.Menu):
    bl_idname = 'MN_MT_NODE_BONDS'
    bl_label = ''
    
    def draw(self, context):
        layout = self.layout
        build_menu(layout, menu_items['bonds'])


class MN_MT_Node_Style(bpy.types.Menu):
    bl_idname = 'MN_MT_NODE_STYLE'
    bl_label = ''
    
    def draw(self, context):
        layout = self.layout
        layout.operator_context = "INVOKE_DEFAULT"
        build_menu(layout, menu_items['style'])

class MN_MT_Node_Select(bpy.types.Menu):
    bl_idname = 'MN_MT_NODE_SELECT'
    bl_label = ''
    
    @classmethod
    def poll(cls, context):
        return True
    
    def draw(self, context):
        layout = self.layout
        layout.operator_context = "INVOKE_DEFAULT"
        build_menu(layout, menu_items['select'])

class MN_MT_Node_Assembly(bpy.types.Menu):
    bl_idname = 'MN_MT_NODE_ASSEMBLY'
    bl_label = ''
    
    @classmethod
    def poll(cls, context):
        return True
    
    def draw(self, context):
        layout = self.layout
        layout.operator_context = "INVOKE_DEFAULT"
        build_menu(layout, menu_items['assembly'])

class MN_MT_Node_DNA(bpy.types.Menu):
    bl_idname = 'MN_MT_NODE_DNA'
    bl_label = ''
    
    def draw(self, context):
        layout = self.layout
        build_menu(layout, menu_items['dna'])

class MN_MT_Node_Animate(bpy.types.Menu):
    bl_idname = 'MN_MT_NODE_ANIMATE'
    bl_label = ''
    
    def draw(self, context):
        layout = self.layout
        build_menu(layout, menu_items['animate'])

class MN_MT_Node_Utils(bpy.types.Menu):
    bl_idname = 'MN_MT_NODE_UTILS'
    bl_label = ''
    
    def draw(self, context):
        build_menu(self.layout, menu_items['utils'])

class MN_MT_Node_CellPack(bpy.types.Menu):
    bl_idname = "MN_MT_NODE_CELLPACK"
    bl_label = ''
    
    def draw(self, context):
        layout = self.layout
        build_menu(layout, menu_items['cellpack'])

class MN_MT_Node_Density(bpy.types.Menu):
    bl_idname = 'MN_MT_NODE_DENSITY'
    bl_label = ''
    
    def draw(self, context):
        layout = self.layout
        layout.operator_context = "INVOKE_DEFAULT"
        build_menu(layout, menu_items['density'])


class MN_MT_Node(bpy.types.Menu):
    bl_idname = "MN_MT_NODE"
    bl_label = "Menu for Adding Nodes in GN Tree"

    @classmethod
    def poll(cls, context):
        return True

    def draw(self, context):
        layout = self.layout.column_flow(columns=1)
        layout.operator_context = "INVOKE_DEFAULT"

        layout.menu('MN_MT_NODE_STYLE', text='Style', icon_value=77)
        layout.menu('MN_MT_NODE_SELECT', text='Selection', icon_value=256)
        layout.menu('MN_MT_NODE_COLOR', text='Color', icon='COLORSET_07_VEC')
        layout.menu('MN_MT_NODE_ANIMATE', text='Animation', icon_value=409)
        layout.menu('MN_MT_NODE_ASSEMBLY', text='Assemblies', icon='GROUP_VERTEX')
        layout.menu('MN_MT_NODE_CELLPACK', text='CellPack', icon='PARTICLE_POINT')
        layout.menu('MN_MT_NODE_DENSITY', text='Density', icon="VOLUME_DATA")
        layout.menu('MN_MT_NODE_DNA', text='DNA', icon='GP_SELECT_BETWEEN_STROKES')
        layout.menu('MN_MT_NODE_UTILS', text='Utilities', icon_value=92)

def MN_add_node_menu(self, context):
    if ('GeometryNodeTree' == bpy.context.area.spaces[0].tree_type):
        layout = self.layout
        layout.menu('MN_MT_NODE', text='Molecular Nodes', icon_value=88)



import bpy
import numpy as np
import warnings
import time

from ..blender import (
    coll, obj
)
from .. import data
from .. import color


bpy.types.Scene.MN_import_centre = bpy.props.BoolProperty(
    name = "Centre Structure", 
    description = "Move the imported Molecule on the World Origin",
    default = False
    )
bpy.types.Scene.MN_import_del_solvent = bpy.props.BoolProperty(
    name = "Remove Solvent", 
    description = "Delete the solvent from the structure on import",
    default = True
    )
bpy.types.Scene.MN_import_panel_selection = bpy.props.IntProperty(
    name = "MN_import_panel_selection", 
    description = "Import Panel Selection", 
    subtype = 'NONE',
    default = 0
)
bpy.types.Scene.MN_import_build_assembly = bpy.props.BoolProperty(
    name = 'Build Assembly', 
    default = False
)

class MolecularNodesObjectProperties(bpy.types.PropertyGroup):
    subframes: bpy.props.IntProperty(
        name = "Subframes", 
        description = "Number of subframes to interpolate for MD trajectories.", 
        default = 0
    )
    molecule_type: bpy.props.StringProperty(
        name = "Molecular Type", 
        description = "How the file was imported, dictating how MN interacts with it.", 
        default = ""
    )
    pdb_code: bpy.props.StringProperty(
        name = "PDB", 
        description = "PDB code used to download this structure.", 
        maxlen = 4,
        options={'HIDDEN'}
    )



def pdb_get_b_factors(file):
    """
    Get a list, which contains a numpy array for each model containing the b-factors.
    """
    b_factors = []
    for model in range(file.get_model_count()):
        atoms = file.get_structure(model = model + 1, extra_fields = ['b_factor'])
        b_factors.append(atoms.b_factor)
    return b_factors

def get_secondary_structure(array, file) -> np.array:
    """
    Gets the secondary structure annotation that is included in mmtf files and returns it as a numerical numpy array.

    Parameters:
    -----------
    array : numpy.array
        The molecular coordinates array, from mmtf.get_structure()
    file : mmtf.MMTFFile
        The MMTF file containing the secondary structure information, from mmtf.MMTFFile.read()

    Returns:
    --------
    atom_sse : numpy.array
        Numerical numpy array representing the secondary structure of the molecule.
    
    Description:
    ------------
    This function uses the biotite.structure package to extract the secondary structure information from the MMTF file.
    The resulting secondary structures are `1: Alpha Helix, 2: Beta-sheet, 3: loop`.
    """
    
    from biotite.structure import spread_residue_wise
    
    sec_struct_codes = {
        -1: "X",
        0 : "I",
        1 : "S",
        2 : "H",
        3 : "E",
        4 : "G",
        5 : "B",
        6 : "T",
        7 : "C"
    }
    
    dssp_to_abc = {
        "X" : 0,
        "I" : 1, #"a",
        "S" : 3, #"c",
        "H" : 1, #"a",
        "E" : 2, #"b",
        "G" : 1, #"a",
        "B" : 2, #"b",
        "T" : 3, #"c",
        "C" : 3 #"c"
    }
    
    try:
        sse = file["secStructList"]
    except KeyError:
        ss_int = np.full(len(array), 3)
        print('Warning: "secStructList" field missing from MMTF file. Defaulting \
            to "loop" for all residues.')
    else:
        ss_int = np.array(
            [dssp_to_abc.get(sec_struct_codes.get(ss)) for ss in sse], 
            dtype = int
        )
    atom_sse = spread_residue_wise(array, ss_int)
    
    return atom_sse


def comp_secondary_structure(array):
    """Use dihedrals to compute the secondary structure of proteins

    Through biotite built-in method derivated from P-SEA algorithm (Labesse 1997)
    Returns an array with secondary structure for each atoms where:
    - 0 = '' = non-protein or not assigned by biotite annotate_sse
    - 1 = a = alpha helix
    - 2 = b = beta sheet
    - 3 = c = coil

    Inspired from https://www.biotite-python.org/examples/gallery/structure/transketolase_sse.html
    """
    #TODO Port [PyDSSP](https://github.com/ShintaroMinami/PyDSSP)
    from biotite.structure import annotate_sse, spread_residue_wise

    conv_sse_char_int = {'a': 1, 'b': 2, 'c': 3, '': 0} 

    char_sse = annotate_sse(array)
    int_sse = np.array([conv_sse_char_int[char] for char in char_sse], dtype=int)
    atom_sse = spread_residue_wise(array, int_sse)
        
    return atom_sse

def create_molecule(array, 
                    name, 
                    centre = False, 
                    file = None,
                    calculate_ss = False,
                    del_solvent = False, 
                    style = 0,
                    collection = None, 
                    verbose = False
                    ):
    import biotite.structure as struc
    
    frames = None
    if isinstance(array, struc.AtomArrayStack):
        if array.stack_depth() > 1:
            frames = array
        array = array[0]
    
    # remove the solvent from the structure if requested
    if del_solvent:
        array = array[np.invert(struc.filter_solvent(array))]

    world_scale = 0.01
    locations = array.coord * world_scale
    
    centroid = np.array([0, 0, 0])
    if centre:
        centroid = struc.centroid(array) * world_scale
    

    # subtract the centroid from all of the positions to localise the molecule on the world origin
    if centre:
        locations = locations - centroid

    if not collection:
        collection = coll.mn()
    
    bonds_array = []
    bond_idx = []

    if array.bonds:
        bonds_array = array.bonds.as_array()
        bond_idx = bonds_array[:, [0, 1]]
        bond_types = bonds_array[:, 2].copy(order = 'C') # the .copy(order = 'C') is to fix a weird ordering issue with the resulting array
    
    mol = obj.create_object(name=name, collection=collection, locations=locations, edges=bond_idx)
    
    # Add information about the bond types to the model on the edge domain
    # Bond types: 'ANY' = 0, 'SINGLE' = 1, 'DOUBLE' = 2, 'TRIPLE' = 3, 'QUADRUPLE' = 4
    # 'AROMATIC_SINGLE' = 5, 'AROMATIC_DOUBLE' = 6, 'AROMATIC_TRIPLE' = 7
    # https://www.biotite-python.org/apidoc/biotite.structure.BondType.html#biotite.structure.BondType
    if array.bonds:
        obj.add_attribute(object = mol, name = 'bond_type', data = bond_types, type = "INT", domain = "EDGE")

    # The attributes for the model are initially defined as single-use functions. This allows
    # for a loop that attempts to add each attibute by calling the function. Only during this
    # loop will the call fail if the attribute isn't accessible, and the warning is reported
    # there rather than setting up a try: except: for each individual attribute which makes
    # some really messy code.
    
    # I still don't like this as an implementation, and welcome any cleaner approaches that 
    # anybody might have.
    
    def att_atomic_number():
        atomic_number = np.array(list(map(
            lambda x: data.elements.get(x, {'atomic_number': -1}).get("atomic_number"), 
            np.char.title(array.element))))
        return atomic_number
    
    def att_res_id():
        return array.res_id
    
    def att_res_name():
        other_res = []
        counter = 0
        id_counter = -1
        res_names = array.res_name
        res_ids = array.res_id
        res_nums  = []
        
        for name in res_names:
            res_num = data.residues.get(name, {'res_name_num': -1}).get('res_name_num')
            
            if res_num == 9999:
                if res_names[counter - 1] != name or res_ids[counter] != res_ids[counter - 1]:
                    id_counter += 1
                
                unique_res_name = str(id_counter + 100) + "_" + str(name)
                other_res.append(unique_res_name)
                
                num = np.where(np.isin(np.unique(other_res), unique_res_name))[0][0] + 100
                res_nums.append(num)
            else:
                res_nums.append(res_num)
            counter += 1

        mol['ligands'] = np.unique(other_res)
        return np.array(res_nums)

    
    def att_chain_id():
        chain_id = np.searchsorted(np.unique(array.chain_id), array.chain_id)
        return chain_id
    
    def att_entity_id():
        return array.entity_id
    
    def att_b_factor():
        return array.b_factor
    
    def att_vdw_radii():
        vdw_radii =  np.array(list(map(
            # divide by 100 to convert from picometres to angstroms which is what all of coordinates are in
            lambda x: data.elements.get(x, {'vdw_radii': 100}).get('vdw_radii', 100) / 100,  
            np.char.title(array.element)
            )))
        return vdw_radii * world_scale
    
    def att_atom_name():
        atom_name = np.array(list(map(
            lambda x: data.atom_names.get(x, -1), 
            array.atom_name
        )))
        
        return atom_name

    def att_lipophobicity():
        lipo = np.array(list(map(
            lambda x, y: data.lipophobicity.get(x, {"0": 0}).get(y, 0),
            array.res_name, array.atom_name
        )))
        
        return lipo
    
    def att_charge():
        charge = np.array(list(map(
            lambda x, y: data.atom_charge.get(x, {"0": 0}).get(y, 0),
            array.res_name, array.atom_name
        )))
        return charge
    
    def att_color():
        return color.color_chains(att_atomic_number(), att_chain_id()).reshape(-1)
    
    def att_is_alpha():
        return np.isin(array.atom_name, 'CA')
    
    def att_is_solvent():
        return struc.filter_solvent(array)
    
    def att_is_backbone():
        """
        Get the atoms that appear in peptide backbone or nucleic acid phosphate backbones.
        Filter differs from the Biotite's `struc.filter_peptide_backbone()` in that this
        includes the peptide backbone oxygen atom, which biotite excludes. Additionally 
        this selection also includes all of the atoms from the ribose in nucleic acids, 
        and the other phosphate oxygens.
        """
        backbone_atom_names = [
            'N', 'C', 'CA', 'O',                    # peptide backbone atoms
            "P", "O5'", "C5'", "C4'", "C3'", "O3'", # 'continuous' nucleic backbone atoms
            "O1P", "OP1", "O2P", "OP2",             # alternative names for phosphate O's
            "O4'", "C1'", "C2'", "O2'"              # remaining ribose atoms
        ]
        
        is_backbone = np.logical_and(
            np.isin(array.atom_name, backbone_atom_names), 
            np.logical_not(struc.filter_solvent(array))
        )
        return is_backbone
    
    def att_is_nucleic():
        return struc.filter_nucleotides(array)
    
    def att_is_peptide():
        aa = struc.filter_amino_acids(array)
        con_aa = struc.filter_canonical_amino_acids(array)
        
        return aa | con_aa
    
    def att_is_hetero():
        return array.hetero
    
    def att_is_carb():
        return struc.filter_carbohydrates(array)

    def att_sec_struct():
        if calculate_ss or not file:
            return comp_secondary_structure(array)
        else:
            return get_secondary_structure(array, file)
    

    

    
    # these are all of the attributes that will be added to the structure
    # TODO add capcity for selection of particular attributes to include / not include to potentially
    # boost performance, unsure if actually a good idea of not. Need to do some testing.
    attributes = (
        {'name': 'res_id',          'value': att_res_id,              'type': 'INT',     'domain': 'POINT'},
        {'name': 'res_name',        'value': att_res_name,            'type': 'INT',     'domain': 'POINT'},
        {'name': 'atomic_number',   'value': att_atomic_number,       'type': 'INT',     'domain': 'POINT'},
        {'name': 'b_factor',        'value': att_b_factor,            'type': 'FLOAT',   'domain': 'POINT'},
        {'name': 'vdw_radii',       'value': att_vdw_radii,           'type': 'FLOAT',   'domain': 'POINT'},
        {'name': 'chain_id',        'value': att_chain_id,            'type': 'INT',     'domain': 'POINT'},
        {'name': 'entity_id',       'value': att_entity_id,           'type': 'INT',     'domain': 'POINT'},
        {'name': 'atom_name',       'value': att_atom_name,           'type': 'INT',     'domain': 'POINT'},
        {'name': 'lipophobicity',   'value': att_lipophobicity,       'type': 'FLOAT',   'domain': 'POINT'},
        {'name': 'charge',          'value': att_charge,              'type': 'FLOAT',   'domain': 'POINT'},
        {'name': 'Color',           'value': att_color,               'type': 'FLOAT_COLOR',   'domain': 'POINT'},
        
        {'name': 'is_backbone',     'value': att_is_backbone,         'type': 'BOOLEAN', 'domain': 'POINT'},
        {'name': 'is_alpha_carbon', 'value': att_is_alpha,            'type': 'BOOLEAN', 'domain': 'POINT'},
        {'name': 'is_solvent',      'value': att_is_solvent,          'type': 'BOOLEAN', 'domain': 'POINT'},
        {'name': 'is_nucleic',      'value': att_is_nucleic,          'type': 'BOOLEAN', 'domain': 'POINT'},
        {'name': 'is_peptide',      'value': att_is_peptide,          'type': 'BOOLEAN', 'domain': 'POINT'},
        {'name': 'is_hetero',       'value': att_is_hetero,           'type': 'BOOLEAN', 'domain': 'POINT'},
        {'name': 'is_carb',         'value': att_is_carb,             'type': 'BOOLEAN', 'domain': 'POINT'},
        {'name': 'sec_struct',      'value': att_sec_struct,          'type': 'INT',     'domain': 'POINT'}
    )
    
    # assign the attributes to the object
    for att in attributes:
        if verbose:
            start = time.process_time()
        try:
            obj.add_attribute(mol, att['name'], att['value'](), att['type'], att['domain'])
            if verbose:
                print(f'Added {att["name"]} after {time.process_time() - start} s')
        except :
            if verbose:
                warnings.warn(f"Unable to add attribute: {att['name']}")
                print(f'Failed adding {att["name"]} after {time.process_time() - start} s')

    if frames:
        try:
            b_factors = pdb_get_b_factors(file)
        except:
            b_factors = None
        
        coll_frames = coll.frames(mol.name, parent = coll.data())
        
        for i, frame in enumerate(frames):
            obj_frame = obj.create_object(
                name = mol.name + '_frame_' + str(i), 
                collection=coll_frames, 
                locations= frame.coord * world_scale - centroid
            )
            if b_factors:
                try:
                    obj.add_attribute(obj_frame, 'b_factor', b_factors[i])
                except:
                    b_factors = False
        
        # disable the frames collection so it is not seen
        # bpy.context.view_layer.layer_collection.children[''].children[coll_frames.name].exclude = True
    else:
        coll_frames = None
    
    mol.mn['molcule_type'] = 'pdb'
    
    # add custom properties to the actual blender object, such as number of chains, biological assemblies etc
    # currently biological assemblies can be problematic to holding off on doing that
    try:
        mol['chain_id_unique'] = list(np.unique(array.chain_id))
    except:
        warnings.warn('No chain information detected.')
    
    try: 
        mol['entity_names'] = [ent['description'] for ent in file['entityList']]
    except:
        pass
    
    return mol, coll_frames






import bpy
import traceback
import os
import zipfile
import numpy as np
from bpy.app.translations import pgettext_tip as tip_

from ..ui.pref import ADDON_DIR


def lerp(a: np.ndarray, b: np.ndarray, t: float = 0.5) -> np.ndarray:
    """
    Linearly interpolate between two values.

    Parameters
    ----------
    a : array_like
        The starting value.
    b : array_like
        The ending value.
    t : float, optional
        The interpolation parameter. Default is 0.5.

    Returns
    -------
    array_like
        The interpolated value(s).

    Notes
    -----
    This function performs linear interpolation between `a` and `b` using the
    interpolation parameter `t` such that the result lies between `a` and `b`.

    Examples
    --------
    >>> lerp(1, 2, 0.5)
    1.5

    >>> lerp(3, 7, 0.2)
    3.8

    >>> lerp([1, 2, 3], [4, 5, 6], 0.5)
    array([2.5, 3.5, 4.5])

    """
    return np.add(a, np.multiply(np.subtract(b, a), t))


def _module_filesystem_remove(path_base, module_name):
    # taken from the bpy.ops.preferences.app_template_install() operator source code
    # Remove all Python modules with `module_name` in `base_path`.
    # The `module_name` is expected to be a result from `_zipfile_root_namelist`.
    import os
    import shutil
    module_name = os.path.splitext(module_name)[0]
    for f in os.listdir(path_base):
        f_base = os.path.splitext(f)[0]
        if f_base == module_name:
            f_full = os.path.join(path_base, f)
            if os.path.isdir(f_full):
                shutil.rmtree(f_full)
            else:
                os.remove(f_full)

def _zipfile_root_namelist(file_to_extract):
    # taken from the bpy.ops.preferences.app_template_install() operator source code
    # Return a list of root paths from zipfile.ZipFile.namelist.
    import os
    root_paths = []
    for f in file_to_extract.namelist():
        # Python's `zipfile` API always adds a separate at the end of directories.
        # use `os.path.normpath` instead of `f.removesuffix(os.sep)`
        # since paths could be stored as `./paths/./`.
        #
        # Note that `..` prefixed paths can exist in ZIP files but they don't write to parent directory when extracting.
        # Nor do they pass the `os.sep not in f` test, this is important,
        # otherwise `shutil.rmtree` below could made to remove directories outside the installation directory.
        f = os.path.normpath(f)
        if os.sep not in f:
            root_paths.append(f)
    return root_paths

def template_install():
    print(os.path.abspath(ADDON_DIR))
    template = os.path.join(os.path.abspath(ADDON_DIR), 'assets', 'template', 'MolecularNodes.zip')
    _install_template(template, 'Molecular Nodes')
    bpy.utils.refresh_script_paths()

def template_uninstall():
    import shutil
    for folder in bpy.utils.app_template_paths():
        path = os.path.join(os.path.abspath(folder), 'MolecularNodes')
        if os.path.exists(path):
            shutil.rmtree(path)
    bpy.utils.refresh_script_paths()

def _install_template(filepath, subfolder = '', overwrite = True):
    # taken from the bpy.ops.preferences.app_template_install() operator source code

    path_app_templates = bpy.utils.user_resource(
        'SCRIPTS',
        path=os.path.join("startup", "bl_app_templates_user", subfolder),
        create=True,
    )

    if not os.path.isdir(path_app_templates):
        try:
            os.makedirs(path_app_templates, exist_ok=True)
        except:
            traceback.print_exc()

    app_templates_old = set(os.listdir(path_app_templates))

    # check to see if the file is in compressed format (.zip)
    if zipfile.is_zipfile(filepath):
        try:
            file_to_extract = zipfile.ZipFile(filepath, 'r')
        except:
            traceback.print_exc()
            return {'CANCELLED'}

        file_to_extract_root = _zipfile_root_namelist(file_to_extract)
        if overwrite:
            for f in file_to_extract_root:
                _module_filesystem_remove(path_app_templates, f)
        else:
            for f in file_to_extract_root:
                path_dest = os.path.join(path_app_templates, os.path.basename(f))
                if os.path.exists(path_dest):
                    # self.report({'WARNING'}, tip_("File already installed to %r\n") % path_dest)
                    return {'CANCELLED'}

        try:  # extract the file to "bl_app_templates_user"
            file_to_extract.extractall(path_app_templates)
        except:
            traceback.print_exc()
            return {'CANCELLED'}

    else:
        # Only support installing zipfiles
        print('no zipfile')
        return {'CANCELLED'}

    app_templates_new = set(os.listdir(path_app_templates)) - app_templates_old

    # in case a new module path was created to install this addon.
    bpy.utils.refresh_script_paths()

    # print message
    msg = (
        tip_("Template Installed (%s) from %r into %r") %
        (", ".join(sorted(app_templates_new)), filepath, path_app_templates)
    )
    print(msg)


import numpy as np
# import biotite.structure.io.mmtf as mmtf
from . import AssemblyParser


class PDBAssemblyParser(AssemblyParser):
    ### Implementation adapted from ``biotite.structure.io.pdb.file``

    def __init__(self, pdb_file):
        self._file = pdb_file
    

    def list_assemblies(self):
        return self._file.list_assemblies()
    

    def get_transformations(self, assembly_id):
        import biotite
        # Get lines containing transformations for assemblies
        remark_lines = self._file.get_remark(350)
        if remark_lines is None:
            raise biotite.InvalidFileError(
                "File does not contain assembly information (REMARK 350)"
            )
        # Get lines corresponding to selected assembly ID
        assembly_start_i = None
        assembly_stop_i = None
        for i, line in enumerate(remark_lines):
            if line.startswith("BIOMOLECULE"):
                current_assembly_id = line[12:].strip()
                if assembly_start_i is not None:
                    # Start was already found -> this is the next entry
                    # -> this is the stop
                    assembly_stop_i = i
                    break
                if current_assembly_id == assembly_id:
                    assembly_start_i = i
        # In case of the final assembly of the file,
        # the 'stop' is the end of REMARK 350 lines
        assembly_stop_i = len(remark_lines) if assembly_stop_i is None else i
        if assembly_start_i is None:
            raise KeyError(
                f"The assembly ID '{assembly_id}' is not found"
            )
        assembly_lines = remark_lines[assembly_start_i : assembly_stop_i]

        # Get transformations for a sets of chains
        transformations = []
        chain_set_start_indices = [
            i for i, line in enumerate(assembly_lines)
            if line.startswith("APPLY THE FOLLOWING TO CHAINS")
        ]
        # Add exclusive stop at end of records
        chain_set_start_indices.append(len(assembly_lines))
        assembly = None
        for i in range(len(chain_set_start_indices) - 1):
            start = chain_set_start_indices[i]
            stop = chain_set_start_indices[i+1]
            # Read affected chain IDs from the following line(s)
            affected_chain_ids = []
            transform_start = None
            for j, line in enumerate(assembly_lines[start : stop]):
                if line.startswith("APPLY THE FOLLOWING TO CHAINS:") or \
                   line.startswith("                   AND CHAINS:"):
                        affected_chain_ids += [
                            chain_id.strip() 
                            for chain_id in line[30:].split(",")
                        ]
                else:
                    # Chain specification has finished
                    # BIOMT lines start directly after chain specification
                    transform_start = start + j
                    break
            # Parse transformations from BIOMT lines
            if transform_start is None:
                raise biotite.InvalidFileError(
                    "No 'BIOMT' records found for chosen assembly"
                )
            rotations, translations = _parse_transformations(
                assembly_lines[transform_start : stop]
            )
            for rotation, translation in zip(rotations, translations):
                transformations.append((
                    np.array(affected_chain_ids, dtype="U4").tolist(),
                    rotation.tolist(),
                    translation.tolist()
                ))

        return transformations
    
    def get_assemblies(self):
        assembly_dict = {}
        for assembly_id in self.list_assemblies():
            assembly_dict[assembly_id] = self.get_transformations(assembly_id)
        
        return assembly_dict


def _parse_transformations(lines):
    """
    Parse the rotation and translation transformations from
    *REMARK* 290 or 350.
    Return as array of matrices and vectors respectively
    """
    import biotite
    # Each transformation requires 3 lines for the (x,y,z) components
    if len(lines) % 3 != 0:
        raise biotite.InvalidFileError("Invalid number of transformation vectors")
    n_transformations = len(lines) // 3

    rotations = np.zeros((n_transformations, 3, 3), dtype=float)
    translations = np.zeros((n_transformations, 3), dtype=float)

    transformation_i = 0
    component_i = 0
    for line in lines:
        # The first two elements (component and
        # transformation index) are not used
        transformations = [float(e) for e in line.split()[2:]]
        if len(transformations) != 4:
            raise biotite.InvalidFileError(
                "Invalid number of transformation vector elements"
            )
        rotations[transformation_i, component_i, :] = transformations[:3]
        translations[transformation_i, component_i] = transformations[3]

        component_i += 1
        if component_i == 3:
            # All (x,y,z) components were parsed
            # -> head to the next transformation 
            transformation_i += 1
            component_i = 0
    
    return rotations, translations


import numpy as np
from . import AssemblyParser

class MMTFAssemblyParser(AssemblyParser):
    ### Implementation adapted from ``biotite.structure.io.mmtf.assembly``

    def __init__(self, mmtf_file):
        self._file = mmtf_file
    

    def list_assemblies(self):
        import biotite.structure.io.mmtf as mmtf
        return mmtf.list_assemblies(self._file)
    

    def get_transformations(self, assembly_id):
        import biotite
        # Find desired assembly
        selected_assembly = None
        if not "bioAssemblyList" in self._file:
            raise biotite.InvalidFileError(
                "File does not contain assembly information "
                "(missing 'bioAssemblyList')"
            )
        for assembly in self._file["bioAssemblyList"]:
            current_assembly_id = assembly["name"]
            transform_list = assembly["transformList"]
            if current_assembly_id == assembly_id:
                selected_assembly = transform_list
                break
        if selected_assembly is None:
            raise KeyError(
                f"The assembly ID '{assembly_id}' is not found"
            )

        # Parse transformations from assembly
        transformations = []
        for transform in selected_assembly:
            matrix = np.array(transform["matrix"]).reshape(4, 4).copy(order = 'C') # order needs to be 'c' otherwise Blender doesn't like it
            chain_ids = np.array(self._file["chainNameList"], dtype="U4")
            affected_chain_ids = chain_ids[transform["chainIndexList"]]
            transformations.append((
                affected_chain_ids.tolist(),
                matrix[:3, :3].tolist(),
                matrix[:3, 3].tolist()
            ))
        
        return transformations
    
    def get_assemblies(self):
        assembly_dict = {}
        for assembly_id in self.list_assemblies():
            assembly_dict[assembly_id] = self.get_transformations(assembly_id)
        
        return assembly_dict


import bpy
import pathlib
from .. import pkg
from bpy.types import AddonPreferences

install_instructions = "https://bradyajohnston.github.io/MolecularNodes/installation.html#installing-biotite-mdanalysis"
ADDON_DIR = pathlib.Path(__file__).resolve().parent.parent

bpy.types.Scene.pypi_mirror_provider = bpy.props.StringProperty(
    name = 'pypi_mirror_provider', 
    description = 'PyPI Mirror Provider', 
    options = {'TEXTEDIT_UPDATE','LIBRARY_EDITABLE'}, 
    default = 'Default', 
    subtype = 'NONE', 
    search = pkg.get_pypi_mirror_alias,
    )


def apple_silicon_warning(layout):
    box = layout.box()
    box.alert = True
    box.label(text = "On M1/M2/M3 macOS, manual installation is required.")
    box.operator(
        "wm.url_open", text = "Installation Instructions", icon = 'HELP'
    ).url = install_instructions
    

# Defines the preferences panel for the addon, which shows the buttons for 
# installing and reinstalling the required python packages defined in 'requirements.txt'
class MolecularNodesPreferences(AddonPreferences):
    bl_idname = 'molecularnodes'

    def draw(self, context):
        layout = self.layout
        layout.label(text = "Install the required packages for MolecularNodes.")
        
        col_main = layout.column(heading = '', align = False)
        row_import = col_main.row()
        row_import.prop(bpy.context.scene, 'pypi_mirror_provider',text='Set PyPI Mirror')
        
        pkgs = pkg.get_pkgs()
        for package in pkgs.values():
            row = layout.row()
            col = row.column()
            row = col.row()
            if pkg._is_apple_silicon and package['name'] == "MDAnalysis":
                apple_silicon_warning(row)
                row = col.row()
                row.enabled = False
                row.alert = True
            pkg.button_install_pkg(
                layout = row, 
                name = package.get('name'), 
                version = package.get('version'), 
                desc = package.get('desc')
                )



import bpy
from ..blender import nodes

def build_menu(layout, items):
    for item in items:
        # print(item)
        if item == "break":
            layout.separator()
        elif item['label'] == "custom":
            for button in item['values']:
                item['function'](layout, 
                    label = button['label'], 
                    field = button['field'], 
                    prefix = button['prefix'], 
                    property_id = button['property_id']
                    )
        elif item['name'].startswith("mn."):
            layout.operator(item['name'])
        else:
            menu_item_interface(layout, item['label'], item['name'], item['description'].removesuffix('.'))

def menu_item_interface(layout_function, 
                        label, 
                        node_name, 
                        node_description='Add custom MolecularNodes node group.', 
                        node_link = False
                        ):
    op = layout_function.operator('mn.add_custom_node_group', text = label)
    op.node_label = label
    op.node_name = node_name
    op.node_description = node_description
    op.node_link = node_link

def button_custom_color(layout, label, field, prefix, property_id, starting_value = 0):
    op = layout.operator('mn.color_custom', text = label)
    op.field = field
    op.prefix = prefix
    op.node_property = property_id
    op.node_name = label.lower()
    op.starting_value = starting_value
    op.description = f"Choose individual colors for each {label}"

def button_custom_selection(layout, label, field, prefix, property_id, starting_value = 0):
    op = layout.operator('mn.selection_custom', text = label)
    op.field = field
    op.prefix = prefix
    op.node_property = property_id
    op.node_name = label.lower()
    op.starting_value = starting_value
    op.description = f"Create individual selections for each {label}"




"""
A subpackage for reading rotation matrices and translation vectors
for biological assemblies from different file formats.

The central functions are `get_transformations_`
"""

from abc import ABCMeta, abstractmethod


class AssemblyParser(metaclass=ABCMeta):

    @abstractmethod
    def list_assemblies(self):
        """
        Return a ``list`` of ``str`` containing the available assembly
        IDs.
        """
    
    @abstractmethod
    def get_transformations(self, assembly_id):
        """
        Parse the necessary transformations for a given
        assembly ID.
        
        Return a ``list`` of transformations for a set of chains
        transformations:

        transformations on sets of chains for this assembly
        |          chain IDs affected by the transformation
        |          |        3x3 rotation matrix
        |          |        |        translation vector 
        |          |        |        |
        list[tuple[ndarray, ndarray, ndarray]]]
        """
        
    @abstractmethod
    def get_assemblies(self):
        """
        Parse all the transformations for each assembly, returning a dictionary of 
        key:value pairs of assembly_id:transformations. The transformations list
        comes from the `get_transformations(assembly_id)` method.
        
        Dictionary of all assemblies
        |     Assembly ID
        |     |   List of transformations to create biological assembly.
        |     |   | 
        dict{'1', list[transformations]}
        
        """


import numpy as np
import bpy
from ..blender import  (
    obj, coll
)

def create_data_object(transforms_array, collection = None, name = 'CellPackModel', world_scale = 0.01, fallback=False):
    obj_data = bpy.data.objects.get(name)
    if obj_data and fallback:
        return obj_data
    
    
    # TODO: check back on this, it was breaking downstream and getting key errors
    # transforms_array = get_transforms_from_dict(transforms_dict)
    
    
    chain_ids = np.unique(transforms_array['chain_id'], return_inverse = True)[1] 
    locations = transforms_array['translation'] * world_scale
    
    if not collection:
        collection = coll.data()
    
    obj_data = obj.create_object(locations, collection=collection, name=name)
    obj.add_attribute(obj_data, 'assembly_rotation', transforms_array['rotation'], 'FLOAT_VECTOR', 'POINT')
    obj.add_attribute(obj_data, 'assembly_id', transforms_array['assembly_id'], 'INT', 'POINT')
    obj.add_attribute(obj_data, 'chain_id', chain_ids, 'INT', 'POINT')
    
    return obj_data

# data types for the np.array that will store per-chain symmetry operations
dtype = [
    ('assembly_id', int),
    ('chain_id',    'U10'),
    ('rotation',    float, 3),
    ('translation', float, 3)
    ]

def get_transforms_from_dict(transforms_dict):
    
    results = None
    for assembly_id, transforms in transforms_dict.items():
        transforms = transforms_from_assemblies(transforms, index = int(assembly_id))
        if results is None:
            results = transforms
        else:
            results = np.concatenate((results, transforms), axis = 0)
    
    # currently also using unique to remove duplicates, TODO: look into duplicates
    results = np.unique(results)
    
    return results

def transforms_from_assemblies(assembly_list, index = 0):
    n_transforms = 0
    for assembly in assembly_list:
        n_transforms += len(assembly[0])
    
    results = np.zeros((n_transforms), dtype = dtype)
    
    current_transform = 0
    for i, assembly in enumerate(assembly_list):
        n_chains = len(assembly[0])
        
        mask = np.array(range(n_chains))
        mask += current_transform
        current_transform += n_chains
        transforms = transform_chains(assembly, index = index)
        results[mask] = transforms
    
    return results

def transform_chains(assembly, index = 0):
    
    chains = assembly[0]
    rotation_matrix = np.array(assembly[1]).reshape((3, 3))
    rotation_euler = rotation_from_matrix(rotation_matrix)
    translation_matrix = np.array(assembly[2])
    
    n = len(chains)
    result = np.zeros((n), dtype = dtype)
    
    for i, chain in enumerate(chains):
        result[i]['assembly_id'] = index
        result[i]['chain_id'] = chain
        result[i]['rotation'] = rotation_euler
        result[i]['translation'] = translation_matrix
        
    return result

def rotation_from_matrix(matrix):
    from scipy.spatial.transform import Rotation as R
    import warnings
    
    # calculate the euler rotation from the rotation matrix
    # Blender is 'xyz' euler rotations. Internally they use matrices / quaternions, but
    # current interfaces for geometry nodes are just eulers
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        return R.from_matrix(matrix).as_euler('xyz')


# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTIBILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

bl_info = {
    "name"        : "molecularnodes",
    "author"      : "Brady Johnston",
    "description" : "Toolbox for molecular animations in Blender & Geometry Nodes.",
    "blender"     : (4, 0, 0),
    "version"     : (4, 0, 0),
    "location"    : "Scene Properties -> Molecular Nodes",
    "warning"     : "",
    "doc_url"     : "https://bradyajohnston.github.io/MolecularNodes/",
    "tracker_url" : "https://github.com/BradyAJohnston/MolecularNodes/issues",
    "category"    : "Import"
}

import bpy
from . import io
from .io.mda import _rejuvenate_universe, _sync_universe
from .ui.node_menu import MN_add_node_menu
from .io.load import MolecularNodesObjectProperties
from . import auto_load
from .util.utils import template_install

auto_load.init()

universe_funcs = [_sync_universe, _rejuvenate_universe]

def register():
    auto_load.register()
    bpy.types.NODE_MT_add.append(MN_add_node_menu)
    bpy.types.Object.mn = bpy.props.PointerProperty(type=MolecularNodesObjectProperties)
    for func in universe_funcs:
        try:
            bpy.app.handlers.load_post.append(func)
        except ValueError as e:
            print(f"Filaed to append {func}, error: {e}.")
    template_install()

def unregister():
    try:
        bpy.types.NODE_MT_add.remove(MN_add_node_menu)
        auto_load.unregister()
        del bpy.types.Object.mn
        for func in universe_funcs:
            try:
                bpy.app.handlers.load_post.remove(func)
            except ValueError as e:
                print(f"Failed to remove {func}, error: {e}.")
    except RuntimeError as e:
        print("Unable to unregister: {e}")

# if __name__ == "main":
#     register()

# # register won't be called when MN is run as a module
bpy.app.handlers.load_post.append(_rejuvenate_universe)
bpy.app.handlers.save_post.append(_sync_universe)



import numpy as np
import itertools

from . import AssemblyParser


class CIFAssemblyParser(AssemblyParser):
    ### Implementation adapted from ``biotite.structure.io.pdbx.convert``

    def __init__(self, mmtf_file):
        self._file = mmtf_file
    

    def list_assemblies(self):
        import biotite.structure.io.pdbx as pdbx    
        return list(pdbx.list_assemblies(self._file).keys())
    

    def get_transformations(self, assembly_id):
        import biotite
        assembly_gen_category = self._file.get_category(
            "pdbx_struct_assembly_gen", expect_looped=True
        )
        if assembly_gen_category is None:
            raise biotite.InvalidFileError(
                "File has no 'pdbx_struct_assembly_gen' category"
            )

        struct_oper_category = self._file.get_category(
            "pdbx_struct_oper_list", expect_looped=True
        )
        if struct_oper_category is None:
            raise biotite.InvalidFileError(
                "File has no 'pdbx_struct_oper_list' category"
            )

        if assembly_id not in assembly_gen_category["assembly_id"]:
            raise KeyError(f"File has no Assembly ID '{assembly_id}'")

        # Extract all possible transformations indexed by operation ID
        transformation_dict = _get_transformations(struct_oper_category)

        # Get necessary transformations and the affected chain IDs
        # NOTE: The chains given here refer to the `label_asym_id` field
        # of the `atom_site` category
        # However, by default `PDBxFile` uses the `auth_asym_id` as
        # chain ID
        transformations = []
        for id, op_expr, asym_id_expr in zip(
            assembly_gen_category["assembly_id"],
            assembly_gen_category["oper_expression"],
            assembly_gen_category["asym_id_list"],
        ):
            # Find the operation expressions for given assembly ID
            # We already asserted that the ID is actually present
            if id == assembly_id:
                operations = _parse_operation_expression(op_expr)
                affected_chain_ids = asym_id_expr.split(",")
                for i, operation in enumerate(operations):
                    rotations = []
                    translations = [] 
                    for op_step in operation:
                        rotation, translation = transformation_dict[op_step]
                        rotations.append(rotation)
                        translations.append(translation)
                    total_rotation, total_translation = _chain_transformations(
                        rotations, translations
                    )
                    transformations.append((
                        np.array(affected_chain_ids, dtype="U4").tolist(),
                        total_rotation.tolist(),
                        total_translation.tolist()
                    ))
        
        return transformations
    
    def get_assemblies(self):
        assembly_dict = {}
        for assembly_id in self.list_assemblies():
            assembly_dict[assembly_id] = self.get_transformations(assembly_id)
        
        return assembly_dict


def _chain_transformations(rotations, translations):
    """
    Get a total rotation/translation transformation by combining
    multiple rotation/translation transformations.
    This is done by intermediately combining rotation matrices and
    translation vectors into 4x4 matrices in the form
    
    |r11 r12 r13 t1|
    |r21 r22 r23 t2|
    |r31 r32 r33 t3|
    |0   0   0   1 |.
    """
    total_matrix = np.identity(4)
    for rotation, translation in zip(rotations, translations):
        matrix = np.zeros((4,4))
        matrix[:3, :3] = rotation
        matrix[:3, 3] = translation
        matrix[3, 3] = 1
        total_matrix = matrix @ total_matrix
    
    return total_matrix[:3, :3], total_matrix[:3, 3]



def _get_transformations(struct_oper):
    """
    Get transformation operation in terms of rotation matrix and
    translation for each operation ID in ``pdbx_struct_oper_list``.
    """
    transformation_dict = {}
    for index, id in enumerate(struct_oper["id"]):
        rotation_matrix = np.array(
            [
                [
                    float(struct_oper[f"matrix[{i}][{j}]"][index])
                    for j in (1, 2, 3)
                ]
                for i in (1, 2, 3)
            ]
        )
        translation_vector = np.array(
            [float(struct_oper[f"vector[{i}]"][index]) for i in (1, 2, 3)]
        )
        transformation_dict[id] = (rotation_matrix, translation_vector)
    return transformation_dict


def _parse_operation_expression(expression):
    """
    Get successive operation steps (IDs) for the given
    ``oper_expression``.
    Form the cartesian product, if necessary.
    """
    # Split groups by parentheses:
    # use the opening parenthesis as delimiter
    # and just remove the closing parenthesis
    expressions_per_step = expression.replace(")", "").split("(")
    expressions_per_step = [e for e in expressions_per_step if len(e) > 0]
    # Important: Operations are applied from right to left
    expressions_per_step.reverse()

    operations = []
    for expr in expressions_per_step:
        if "-" in expr:
            if "," in expr:
                for gexpr in expr.split(","):
                    if "-" in gexpr:
                        first, last = gexpr.split("-")
                        operations.append(
                            [str(id) for id in range(int(first), int(last) + 1)]
                        )
                    else:
                        operations.append([gexpr])
            else:
                # Range of operation IDs, they must be integers
                first, last = expr.split("-")
                operations.append(
                    [str(id) for id in range(int(first), int(last) + 1)]
                )
        elif "," in expr:
            # List of operation IDs
            operations.append(expr.split(","))
        else:
            # Single operation ID
            operations.append([expr])

    # Cartesian product of operations
    return list(itertools.product(*operations))



import bpy
import os
import numpy as np
import math
import warnings
from .. import color
from .. import pkg
from ..blender import obj

socket_types = {
        'BOOLEAN'   : 'NodeSocketBool', 
        'GEOMETRY'  : 'NodeSocketGeometry', 
        'INT'       : 'NodeSocketInt', 
        'MATERIAL'  : 'NodeSocketMaterial', 
        'VECTOR'    : 'NodeSocketVector', 
        'STRING'    : 'NodeSocketString', 
        'VALUE'     : 'NodeSocketFloat', 
        'COLLECTION': 'NodeSocketCollection', 
        'TEXTURE'   : 'NodeSocketTexture', 
        'COLOR'     : 'NodeSocketColor', 
        'IMAGE'     : 'NodeSocketImage'
    }

# current implemented representations
styles_mapping = {
    "presets"       : "MN_style_presets",
    'preset_1'      : ".MN_style_preset_1",
    'preset_2'      : ".MN_style_preset_2",
    'preset_3'      : ".MN_style_preset_3",
    'preset_4'      : ".MN_style_preset_4",
    'atoms'         : 'MN_style_spheres',
    'spheres'       : 'MN_style_spheres',
    'vdw'           : 'MN_style_spheres',
    'sphere'        : 'MN_style_spheres',
    'cartoon'       : 'MN_style_cartoon',
    'ribbon'        : 'MN_style_ribbon',
    'surface'       : 'MN_style_surface',
    'ball_and_stick': 'MN_style_ball_and_stick',
    'ball+stick'    : 'MN_style_ball_and_stick',
    'oxdna'         : 'MN_oxdna_style_ribbon', 
    "density_surface": "MN_density_style_surface", 
    "density_wire" : "MN_density_style_wire"
}

STYLE_ITEMS = (
        ('presets', 'Presets', 'A pre-made combination of different styles'),
        ("spheres", "Spheres", "Space-filling atoms style."), 
        ("surface", "Surface", "Solvent-accsible surface."),
        ("cartoon", "Cartoon", "Secondary structure cartoons"), 
        ("ribbon", "Ribbon", "Continuous backbone ribbon."), 
        ("ball_and_stick", "Ball and Stick", "Spheres for atoms, sticks for bonds")
    )

bpy.types.Scene.MN_import_style = bpy.props.EnumProperty(
    name = "Style", 
    description = "Default style for importing molecules.", 
    items = STYLE_ITEMS
)


MN_DATA_FILE = os.path.join(pkg.ADDON_DIR, 'assets', 'MN_data_file.blend')

def inputs(node):
    items = {}
    for item in node.interface.items_tree:
        if item.item_type == "SOCKET":
            if item.in_out == "INPUT":
                items[item.name] = item 
    return items

def outputs(node):
    items = {}
    for item in node.interface.items_tree:
        if item.item_type == "SOCKET":
            if item.in_out == "OUTPUT":
                items[item.name] = item
    return items

def get_output_type(node, type = "INT"):
    for output in node.outputs:
        if output.type == type:
            return output

def set_selection(group, node, selection):
    pos = node.location
    pos = [pos[0] - 200, pos[1] - 200]
    selection.location = pos
    group.links.new(selection.outputs[0], node.inputs['Selection'])
    
    return selection

def add_selection(group, sel_name, input_list, attribute = 'chain_id'):
    style = style_node(group)
    sel_node = add_custom(group, chain_selection('selection', input_list, attribute=attribute).name)
    
    set_selection(group, style, sel_node)
    return sel_node

def get_output(group):
    return group.nodes[bpy.app.translations.pgettext_data("Group Output",)]

def get_input(group):
    return group.nodes[bpy.app.translations.pgettext_data("Group Input",)]

def get_mod(object):
    node_mod = object.modifiers.get('MolecularNodes')
    if not node_mod:
        node_mod = object.modifiers.new("MolecularNodes", "NODES")
    object.modifiers.active = node_mod
    return node_mod

def format_node_name(name):
    "Formats a node's name for nicer printing."
    return name.strip("MN_").replace("_", " ").title()

def  get_nodes_last_output(group):
    output = get_output(group)
    last = output.inputs[0].links[0].from_node
    return last, output

def previous_node(node):
    "Get the node which is the first connection to the first input of this node"
    prev = node.inputs[0].links[0].from_node
    return prev

def style_node(group):
    prev = previous_node(get_output(group))
    is_style_node = ("style" in prev.name)
    while not is_style_node:
        print(prev.name)
        prev = previous_node(prev)
        is_style_node = ("style" in prev.name)
    return prev

def get_style_node(object):
    "Walk back through the primary node connections until you find the first style node"
    group = object.modifiers['MolecularNodes'].node_group
    return style_node(group)

def get_color_node(object):
    "Walk back through the primary node connections until you find the first style node"
    group = object.modifiers['MolecularNodes'].node_group
    for node in group.nodes:
        if node.name == "MN_color_attribute_random":
            return node

def insert_last_node(group, node):
    last, output = get_nodes_last_output(group)
    link = group.links.new
    location = output.location
    output.location = [location[0] + 300, location[1]]
    node.location = [location[0] - 300, location[1]]
    link(last.outputs[0], node.inputs[0])
    link(node.outputs[0], output.inputs[0])

def realize_instances(obj):
    group = obj.modifiers['MolecularNodes'].node_group
    realize = group.nodes.new('GeometryNodeRealizeInstances')
    insert_last_node(group, realize)

def append(node_name, link = False):
    node = bpy.data.node_groups.get(node_name)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        if not node or link:
            bpy.ops.wm.append(
                'EXEC_DEFAULT',
                directory = os.path.join(MN_DATA_FILE, 'NodeTree'), 
                filename = node_name, 
                link = link
            )
    
    return bpy.data.node_groups[node_name]

def MN_base_material():
    """
    Append MN_atomic_material to the .blend file it it doesn't already exist, 
    and return that material.
    """
    
    mat_name = 'MN_atomic_material'
    mat = bpy.data.materials.get(mat_name)
    
    if not mat:
        print('appending material')
        bpy.ops.wm.append(
            directory = os.path.join(MN_DATA_FILE, 'Material'), 
            filename = 'MN_atomic_material', 
            link = False
        )
    
    return bpy.data.materials[mat_name]

def new_group(name = "Geometry Nodes", geometry = True, fallback=True):
    group = bpy.data.node_groups.get(name)
    # if the group already exists, return it and don't create a new one
    if group and fallback:
        return group
    
    # create a new group for this particular name and do some initial setup
    group = bpy.data.node_groups.new(name, 'GeometryNodeTree')
    input_node = group.nodes.new('NodeGroupInput')
    output_node = group.nodes.new('NodeGroupOutput')
    input_node.location.x = -200 - input_node.width
    output_node.location.x = 200
    if geometry:
        group.interface.new_socket('Geometry', in_out='INPUT', socket_type='NodeSocketGeometry')
        group.interface.new_socket('Geometry', in_out='OUTPUT', socket_type='NodeSocketGeometry')
        group.links.new(output_node.inputs[0], input_node.outputs[0])
    return group

def assign_material(node, material = 'default'):
    material_socket = node.inputs.get('Material')
    if material_socket:
        if not material:
            pass
        elif material == "default":
            material_socket.default_value = MN_base_material()
        else:
            material_socket.default_value = material

def add_node(node_name, label: str = '', show_options = False, material = "default"):
    # intended to be called upon button press in the node tree
    
    prev_context = bpy.context.area.type
    bpy.context.area.type = 'NODE_EDITOR'
    # actually invoke the operator to add a node to the current node tree
    # use_transform=True ensures it appears where the user's mouse is and is currently 
    # being moved so the user can place it where they wish
    bpy.ops.node.add_node(
        'INVOKE_DEFAULT', 
        type='GeometryNodeGroup', 
        use_transform=True
        )
    bpy.context.area.type = prev_context
    node = bpy.context.active_node
    node.node_tree = bpy.data.node_groups[node_name]
    node.width = 200.0
    node.show_options = show_options
    
    if label == '':
        node.label = format_node_name(node_name)
    else:
        node.label = label
    node.name = node_name
    
    # if added node has a 'Material' input, set it to the default MN material
    assign_material(node, material=material)

def add_custom(
        group, 
        name, 
        location = [0,0], 
        width = 200,
        material = "default",
        show_options = False, 
        link = False
    ):
    
    node = group.nodes.new('GeometryNodeGroup')
    node.node_tree = append(name, link = link)
    
    # if there is an input socket called 'Material', assign it to the base MN material
    # if another material is supplied, use that instead.
    assign_material(node, material=material)
    
    # move and format the node for arranging
    node.location = location
    node.width = width
    node.show_options = show_options
    node.name = name
    node.label = format_node_name(name)
    
    return node

def change_style_node(object, style):
    group = get_mod(object).node_group
    link = group.links.new
    node_style = get_style_node(object)
    socket_from = node_style.inputs[0].links[0].from_socket
    socket_to   = node_style.outputs[0].links[0].to_socket
    new_style = append(styles_mapping[style])
    node_style.node_tree = new_style
    node_style.name = new_style.name
    node_style.label = format_node_name(new_style.name)
    link(socket_from, node_style.inputs[0])
    link(node_style.outputs[0], socket_to)
    assign_material(get_style_node(object))

def create_starting_nodes_starfile(object):
    # ensure there is a geometry nodes modifier called 'MolecularNodes' that is created and applied to the object
    node_mod = get_mod(object)
    
    node_name = f"MN_starfile_{object.name}"
    
    # if node tree already exists by this name, set it and return it
    group = bpy.data.node_groups.get(node_name)
    if group:
        node_mod.node_group = group
        return group
    
    # create a new GN node group, specific to this particular molecule
    group = new_group(node_name)
    node_mod.node_group = group
    
    group.interface.new_socket('Molecule', in_out='INPUT', socket_type='NodeSocketObject')
    group.interface.new_socket('Image', in_out='INPUT', socket_type='NodeSocketInt')
    group.interface.items_tree["Image"].default_value = 1
    group.interface.items_tree["Image"].min_value = 1
    group.interface.new_socket('Simplify', in_out='INPUT', socket_type='NodeSocketBool')
    
    # move the input and output nodes for the group
    node_input = get_input(group)
    node_input.location = [0, 0]
    node_output = get_output(group)
    node_output.location = [900, 0]

    node_delete = group.nodes.new("GeometryNodeDeleteGeometry")
    node_delete.location = [500, 0]

    node_geom_to_instance = group.nodes.new("GeometryNodeInstanceOnPoints")
    node_geom_to_instance.location = [675, 0]

    node_get_imageid = group.nodes.new("GeometryNodeInputNamedAttribute")
    node_get_imageid.location = [0, 200]
    node_get_imageid.inputs['Name'].default_value = "MOLImageId"
    node_get_imageid.data_type = "INT"

    node_subtract = group.nodes.new("ShaderNodeMath")
    node_subtract.location = [160, 200]
    node_subtract.operation = "SUBTRACT"
    node_subtract.inputs[1].default_value = 1
    node_subtract.inputs[0].default_value = 1


    node_compare = group.nodes.new("FunctionNodeCompare")
    node_compare.location = [320, 200]
    node_compare.operation = "NOT_EQUAL"
    node_compare.data_type = "INT"

    node_object_info = group.nodes.new("GeometryNodeObjectInfo")
    node_object_info.location = [200, -200]

    node_get_rotation = group.nodes.new("GeometryNodeInputNamedAttribute")
    node_get_rotation.location = [450, -200]
    node_get_rotation.inputs['Name'].default_value = "MOLRotation"
    node_get_rotation.data_type = "FLOAT_VECTOR"

    node_get_id = group.nodes.new("GeometryNodeInputID")
    node_get_id.location = [0, -200]

    node_statistics = group.nodes.new("GeometryNodeAttributeStatistic")
    node_statistics.location = [200, -400]

    node_compare_maxid = group.nodes.new("FunctionNodeCompare")
    node_compare_maxid.location = [400, -400]
    node_compare_maxid.operation = "EQUAL"

    node_bool_math = group.nodes.new("FunctionNodeBooleanMath")
    node_bool_math.location = [600, -400]
    node_bool_math.operation = "OR"

    node_switch = group.nodes.new("GeometryNodeSwitch")
    node_switch.location = [800, -400]

    node_cone = group.nodes.new("GeometryNodeMeshCone")
    node_cone.location = [1000, -400]

    link = group.links.new

    link(node_input.outputs[0], node_delete.inputs[0])
    link(node_delete.outputs[0], node_geom_to_instance.inputs[0])
    link(node_geom_to_instance.outputs[0], node_output.inputs[0])

    link(node_input.outputs[1], node_object_info.inputs[0])
    link(node_input.outputs[2], node_subtract.inputs[0])
    link(node_input.outputs[3], node_bool_math.inputs[0])

    link(node_subtract.outputs[0], node_compare.inputs[2])
    link(node_get_imageid.outputs[4], node_compare.inputs[3])
    link(node_compare.outputs[0], node_delete.inputs[1])
    link(node_statistics.outputs[4], node_compare_maxid.inputs[0])
    link(node_compare_maxid.outputs[0], node_bool_math.inputs[1])
    link(node_get_id.outputs[0], node_statistics.inputs[2])
    link(node_object_info.outputs["Geometry"], node_statistics.inputs[0])
    link(node_bool_math.outputs[0], node_switch.inputs[1])
    link(node_object_info.outputs["Geometry"], node_switch.inputs[14])
    link(node_cone.outputs[0], node_switch.inputs[15])
    link(node_switch.outputs[6],     node_geom_to_instance.inputs["Instance"])
    link(node_get_rotation.outputs[0], node_geom_to_instance.inputs["Rotation"])


    # Need to manually set Image input to 1, otherwise it will be 0 (even though default is 1)
    node_mod['Input_3'] = 1

def create_starting_nodes_density(object, threshold = 0.8, style = 'surface'):
    # ensure there is a geometry nodes modifier called 'MolecularNodes' that is created and applied to the object
    mod = get_mod(object)
    node_name = f"MN_density_{object.name}"
    
    # create a new GN node group, specific to this particular molecule
    group = new_group(node_name)
    link = group.links.new
    mod.node_group = group
    
    # move the input and output nodes for the group
    node_input = get_input(group)
    node_input.location = [0, 0]
    node_output = get_output(group)
    node_output.location = [800, 0]
    
    node_density = add_custom(group, styles_mapping[style], [400, 0])
    node_density.inputs['Threshold'].default_value = threshold
    
    link(node_input.outputs[0], node_density.inputs[0])
    link(node_density.outputs[0], node_output.inputs[0])


def create_starting_node_tree(object, coll_frames=None, style="spheres", name=None, set_color=True):
    """
    Create a starting node tree for the inputted object.

    Parameters
    ----------
    object : bpy.types.Object
        Object to create the node tree for.
    coll_frames : bpy.data.collections, optional
        If None, no animation will be created.
        The default is None.
    style : str, optional
        Starting style for the node tree. The default is "spheres".
        Available options are stored as the keys of styles_mapping.
    name : str, optional
        Name of the node tree. If None, a default name will be generated.
        The default is None.
    set_color : bool, optional
        Whether to set up nodes for generating colors in the node tree.
        The default is True.
    """
    # ensure there is a geometry nodes modifier called 'MolecularNodes' that is created and applied to the object
    mod = get_mod(object)
    
    if not name:
        name = f"MN_{object.name}"
    
    # create a new GN node group, specific to this particular molecule
    group = new_group(name)
    link = group.links.new
    mod.node_group = group
    
    # move the input and output nodes for the group
    node_input = get_input(group)
    node_output = get_output(group)
    node_input.location = [0, 0]
    node_output.location = [700, 0]
    
    node_style = add_custom(group, styles_mapping[style], [450, 0])
    link(node_style.outputs[0], node_output.inputs[0])
    link(node_input.outputs[0], node_style.inputs[0])
    
    # if requested, setup the nodes for generating colors in the node tree
    if set_color:
        node_color_set = add_custom(group, 'MN_color_set', [200, 0])
        node_color_common = add_custom(group, 'MN_color_common', [-50, -150])
        node_random_color = add_custom(group, 'MN_color_attribute_random', [-300, -150])
        
        link(node_input.outputs['Geometry'], node_color_set.inputs[0])
        link(node_random_color.outputs['Color'], node_color_common.inputs['Carbon'])
        link(node_color_common.outputs[0], node_color_set.inputs['Color'])    
        link(node_color_set.outputs[0], node_style.inputs[0])
        to_animate = node_color_set
    else:
        to_animate = node_input

    # if multiple frames, set up the required nodes for an animation
    if coll_frames:
        node_output.location = [1100, 0]
        node_style.location = [800, 0]
        
        node_animate_frames = add_custom(group, 'MN_animate_frames', [500, 0])
        node_animate_frames.inputs['Frames'].default_value = coll_frames
        
        node_animate = add_custom(group, 'MN_animate_value', [500, -300])
        link(to_animate.outputs[0], node_animate_frames.inputs[0])
        link(node_animate_frames.outputs[0], node_style.inputs[0])
        link(node_animate.outputs[0], node_animate_frames.inputs['Frame'])


def combine_join_geometry(group, node_list, output = 'Geometry', join_offset = 300):
    link = group.links.new
    max_x = max([node.location[0] for node in node_list])
    node_to_instances = group.nodes.new('GeometryNodeJoinGeometry')
    node_to_instances.location = [int(max_x + join_offset), 0]

    for node in reversed(node_list):
        link(node.outputs[output], node_to_instances.inputs['Geometry'])
    return node_to_instances


def split_geometry_to_instances(name, iter_list=('A', 'B', 'C'), attribute='chain_id'):
    """Create a Node to Split Geometry by an Attribute into Instances
    
    Splits the inputted geometry into instances, based on an attribute field. By
    default this field is the `chain_id` but this can be selected for any field.
    Will loop over each item of the list, so a list of arbitrary items that will 
    define how many times to create the required nodes.
    
    """
    group = new_group(name)
    node_input = get_input(group)
    node_output = get_output(group)

    named_att = group.nodes.new('GeometryNodeInputNamedAttribute')
    named_att.location = [-200, -200]
    named_att.data_type = 'INT'
    named_att.inputs[0].default_value = attribute

    link = group.links.new
    list_sep = []

    for i, chain in enumerate(iter_list):
        pos = [i % 10, math.floor(i / 10)]

        node_split = add_custom(group, '.MN_utils_split_instance')
        node_split.location = [int(250 * pos[0]), int(-300 * pos[1])]
        node_split.inputs['Group ID'].default_value = i
        link(named_att.outputs[4], node_split.inputs['Field'])
        link(node_input.outputs['Geometry'], node_split.inputs['Geometry'])
        list_sep.append(node_split)

    node_instance = combine_join_geometry(group, list_sep, 'Instance')
    node_output.location = [int(10 * 250 + 400), 0]
    link(node_instance.outputs[0], node_output.inputs[0])
    return group


def create_assembly_node_tree(name, iter_list, data_object):
    
    node_group_name = f"MN_assembly_{name}"
    group = new_group(name = node_group_name)
    link = group.links.new
    
    n_assemblies = len(np.unique(obj.get_attribute(data_object, 'assembly_id')))
    
    node_group_instances = split_geometry_to_instances(
        name = f".MN_utils_split_{name}", 
        iter_list = iter_list, 
        attribute = 'chain_id'
    )
    
    node_group_assembly_instance = append('.MN_assembly_instance_chains')
    node_instances = add_custom(group, node_group_instances.name, [0, 0])
    node_assembly = add_custom(group, node_group_assembly_instance.name, [200, 0])
    node_assembly.inputs['data_object'].default_value = data_object
    
    out_sockets = outputs(group)
    out_sockets[list(out_sockets)[0]].name = "Instances"
    
    socket_info = (
        {"name" : "Rotation",    "type": "NodeSocketFloat", "min": 0, "max": 1, "default": 1},
        {"name" : "Translation", "type": "NodeSocketFloat", "min": 0, "max": 1, "default": 1},
        {"name" : "assembly_id", "type": "NodeSocketInt", "min": 1, "max": n_assemblies, "default": 1}
    )
    
    for info in socket_info:
        socket = group.interface.items_tree.get(info['name'])
        if not socket:
            socket = group.interface.new_socket(info['name'], in_out='INPUT', socket_type=info['type'])
        socket.default_value = info['default']
        socket.min_value = info['min']
        socket.max_value = info['max']
        
        link(get_input(group).outputs[info['name']], node_assembly.inputs[info['name']])
    
    get_output(group).location = [400, 0]
    link(get_input(group).outputs[0], node_instances.inputs[0])
    link(node_instances.outputs[0], node_assembly.inputs[0])
    link(node_assembly.outputs[0], get_output(group).inputs[0])
    
    return group


def add_inverse_selection(group):
    output = get_output(group)
    if 'Inverted' not in output.inputs.keys():
        group.interface.new_socket('Inverted', in_out='OUTPUT', socket_type='NodeSocketBool')
    
    loc = output.location
    bool_math = group.nodes.new("FunctionNodeBooleanMath")
    bool_math.location = [loc[0], -100]
    bool_math.operation = "NOT"
    
    group.links.new(output.inputs['Selection'].links[0].from_socket, bool_math.inputs[0])
    group.links.new(bool_math.outputs[0], output.inputs['Inverted'])

def chain_selection(name, input_list, attribute = 'chain_id', starting_value = 0, label_prefix = ""):
    """
    Given a an input_list, will create a node which takes an Integer input, 
    and has a boolean tick box for each item in the input list. The outputs will
    be the resulting selection and the inversion of the selection.
    Can contain a prefix for the resulting labels. Mostly used for constructing 
    chain selections when required for specific proteins.
    """
    # just reutn the group name early if it already exists
    group = bpy.data.node_groups.get(name)
    if group:
        return group
    
    group = new_group(name, geometry=False)
    link = group.links.new
    # create a named attribute node that gets the chain_number attribute
    # and use this for the selection algebra that happens later on
    node_attribute = group.nodes.new("GeometryNodeInputNamedAttribute")
    node_attribute.data_type = 'INT'
    node_attribute.location = [-200, 200]
    node_attribute.inputs[0].default_value = attribute
    node_attribute.outputs.get('Attribute')
    
    # distance horizontally to space all of the created nodes
    node_sep_dis = 180
    previous_node = None
    att_output = get_output_type(node_attribute, 'INT')
    
    for i, chain_name in enumerate(input_list):
        group.interface.new_socket(str(label_prefix) + str(chain_name), in_out='INPUT', socket_type='NodeSocketBool')
        current_node = group.nodes.new("GeometryNodeGroup")
        current_node.node_tree = append('.MN_utils_bool_chain')
        current_node.location = [i * node_sep_dis, 200]
        current_node.inputs["number_matched"].default_value = i + starting_value

        # link from the the named attribute node chain_number into the other inputs
        link(get_input(group).outputs[i], current_node.inputs["bool_include"])
        if previous_node:
            link(previous_node.outputs['number_chain_out'], current_node.inputs['number_chain_in'])
            link(previous_node.outputs['bool_chain_out'], current_node.inputs['bool_chain_in'])
        else:
            link(att_output, current_node.inputs['number_chain_in'])
        previous_node = current_node

    group.interface.new_socket('Selection', in_out='OUTPUT', socket_type='NodeSocketBool')
    group_out = get_output(group)
    group_out.location = [len(input_list) * node_sep_dis, 200]
    link(current_node.outputs['bool_chain_out'], group_out.inputs['Selection'])
    add_inverse_selection(group)

    return group

def chain_color(name, input_list, label_prefix = "Chain ", field = "chain_id", starting_value = 0):
    """
    Given the input list of chain names, will create a node group which uses
    the chain_id named attribute to manually set the colours for each of the chains.
    """
    
    group = new_group(name)
    link = group.links.new
    
    # create a named attribute node that gets the chain_number attribute
    # and use this for the selection algebra that happens later on
    node_att = group.nodes.new("GeometryNodeInputNamedAttribute")
    node_att.data_type = 'INT'
    node_att.location = [-200, 400]
    node_att.inputs[0].default_value = field
    node_att.outputs.get('Attribute')
    
    node_input = get_input(group)
    
    # shortcut for creating new nodes
    new_node = group.nodes.new
    # distance horizontally to space all of the created nodes
    node_sep_dis = 180
    att_output = get_output_type(node_att, 'INT')
    node_color_previous = None
    
    for i, chain_name in enumerate(input_list):
        i += starting_value
        offset = i * node_sep_dis
        current_chain = f"{label_prefix}{chain_name}"
        
        # node compare inputs 2 & 3
        node_compare = new_node('FunctionNodeCompare')
        node_compare.data_type = 'INT'
        node_compare.location = [offset, 100]
        node_compare.operation = 'EQUAL'
        
        node_compare.inputs[3].default_value = i
        
        # link the named attribute to the compare
        link(att_output, node_compare.inputs[2])
        
        node_color = new_node('GeometryNodeSwitch')
        node_color.input_type = 'RGBA'
        node_color.location = [offset, -100]
        
        # create an input for this chain
        group.interface.new_socket(current_chain, in_out='INPUT', socket_type='NodeSocketColor').default_value = color.random_rgb(i)
        # switch input colours 10 and 11
        link(node_input.outputs[current_chain], node_color.inputs[11])
        link(node_compare.outputs['Result'], node_color.inputs['Switch'])
        
        
        if node_color_previous:
            link(node_color_previous.outputs[4], node_color.inputs[10])
        node_color_previous = node_color
    
    group.interface.new_socket('Color', in_out='OUTPUT', socket_type='NodeSocketColor')
    node_output = group.nodes.new("NodeGroupOutput")
    node_output.location = [offset, 200]
    link(node_color.outputs[4], node_output.inputs['Color'])
    
    return group

def resid_multiple_selection(node_name, input_resid_string):
    """
    Returns a node group that takes an integer input and creates a boolean 
    tick box for each item in the input list. Outputs are the selected 
    residues and the inverse selection. Used for constructing chain 
    selections in specific proteins.
    """
        
    #print(f'recieved input: {input_resid_string}')
    # do a cleanning of input string to allow fuzzy input from users
    for c in ";/+ .":
        if c in input_resid_string:
            input_resid_string=input_resid_string.replace(c, ',')

    for c in "_=:":
        if c in input_resid_string:
            input_resid_string=input_resid_string.replace(c, '-')

    #print(f'fixed input:{input_resid_string}')

    # parse input_resid_string into sub selecting string list
    sub_list=[item for item in input_resid_string.split(',') if item]
    
    # distance vertical to space all of the created nodes
    node_sep_dis = -100

    # get the active object, might need to change to taking an object as an input
    # and making it active isntead, to be more readily applied to multiple objects
    
    # create the custom node group data block, where everything will go
    # also create the required group node input and position it
    residue_id_group = bpy.data.node_groups.new(node_name, "GeometryNodeTree")
    node_input = residue_id_group.nodes.new("NodeGroupInput")
    node_input.location = [0, node_sep_dis * len(sub_list)/2]
    
    group_link = residue_id_group.links.new
    new_node = residue_id_group.nodes.new
    
    for residue_id_index,residue_id in enumerate(sub_list):
        
        # add an new node of Select Res ID or MN_sek_res_id_range
        current_node = new_node("GeometryNodeGroup")

        # add an bool_math block 
        bool_math = new_node("FunctionNodeBooleanMath")
        bool_math.location = [400,(residue_id_index+1) * node_sep_dis]
        bool_math.operation = "OR"

        if '-' in residue_id:
            # set two new inputs
            current_node.node_tree = append('MN_select_res_id_range')
            [resid_start, resid_end] = residue_id.split('-')[:2]
            socket_1 = residue_id_group.interface.new_socket('res_id: Min', in_out='INPUT', socket_type='NodeSocketInt')
            socket_1.default_value = int(resid_start)
            socket_2 = residue_id_group.interface.new_socket('res_id: Max', in_out='INPUT', socket_type='NodeSocketInt')
            socket_2.default_value = int(resid_end)
            
            # a residue range
            group_link(node_input.outputs[socket_1.identifier], current_node.inputs[0])
            group_link(node_input.outputs[socket_2.identifier], current_node.inputs[1])
        else:
            # create a node
            current_node.node_tree = append('MN_select_res_id_single')
            socket = residue_id_group.interface.new_socket('res_id', in_out='INPUT', socket_type='NodeSocketInt')
            socket.default_value  = int(residue_id)
            group_link(node_input.outputs[socket.identifier], current_node.inputs[0])
        
        # set the coordinates
        current_node.location = [200,(residue_id_index+1) * node_sep_dis]
        prev = None
        if not prev:
            # link the first residue selection to the first input of its OR block
            group_link(current_node.outputs['Selection'],bool_math.inputs[0])
        else:
            # if it is not the first residue selection, link the output to the previous or block
            group_link(current_node.outputs['Selection'], prev.inputs[1])
        
            # link the ouput of previous OR block to the current OR block
            group_link(prev.outputs[0], bool_math.inputs[0])
        prev = bool_math

    # add a output block
    residue_id_group_out = new_node("NodeGroupOutput")
    residue_id_group_out.location = [800,(residue_id_index + 1) / 2 * node_sep_dis]
    residue_id_group.interface.new_socket('Selection', in_out='OUTPUT', socket_type='NodeSocketBool')
    residue_id_group.interface.new_socket('Inverted', in_out='OUTPUT', socket_type='NodeSocketBool')
    group_link(prev.outputs[0], residue_id_group_out.inputs['Selection'])
    invert_bool_math = new_node("FunctionNodeBooleanMath")
    invert_bool_math.location = [600,(residue_id_index+1)/ 3 * 2 * node_sep_dis]
    invert_bool_math.operation = "NOT"
    group_link(prev.outputs[0], invert_bool_math.inputs[0])
    group_link(invert_bool_math.outputs[0], residue_id_group_out.inputs['Inverted'])
    return residue_id_group


import bpy
from pathlib import Path
import numpy as np

from .load import create_molecule
from ..blender import nodes
from .. import assembly

bpy.types.Scene.MN_pdb_code = bpy.props.StringProperty(
    name = 'PDB', 
    description = 'The 4-character PDB code to download', 
    options = {'TEXTEDIT_UPDATE'}, 
    maxlen = 4
    )
bpy.types.Scene.MN_cache_dir = bpy.props.StringProperty(
    name = 'Cache',
    description = 'Location to cache PDB files',
    options = {'TEXTEDIT_UPDATE'},
    default = str(Path('~', '.MolecularNodes').expanduser()),
    subtype = 'FILE_PATH'
)
bpy.types.Scene.MN_cache = bpy.props.BoolProperty(
    name = "Cache", 
    default = True
)

def load(
    pdb_code,             
    style = 'spheres',               
    centre = False,               
    del_solvent = True,               
    setup_nodes = True,
    cache_dir = None,
    build_assembly = False
    ):
    from biotite import InvalidFileError
    mol, file = open_structure_rcsb(
        pdb_code = pdb_code, 
        cache_dir = cache_dir
        )
    
    if build_assembly:
        centre = False
    
    mol, coll_frames = create_molecule(
        array = mol,
        name = pdb_code,
        file = file,
        calculate_ss = False,
        centre = centre,
        del_solvent = del_solvent, 
        )
    
    mol.mn['pdb_code'] = pdb_code
    mol.mn['molecule_type'] = 'pdb'
    
    if setup_nodes:
        nodes.create_starting_node_tree(
            object = mol, 
            coll_frames=coll_frames, 
            style = style
            )
    
    # mol['bio_transform_dict'] = file['bioAssemblyList']
    
    
    try:
        parsed_assembly_file = assembly.mmtf.MMTFAssemblyParser(file)
        mol['biological_assemblies'] = parsed_assembly_file.get_assemblies()
    except InvalidFileError:
        pass
    
    if build_assembly:
        obj = mol
        transforms_array = assembly.mesh.get_transforms_from_dict(obj['biological_assemblies'])
        data_object = assembly.mesh.create_data_object(
            transforms_array = transforms_array, 
            name = f"data_assembly_{obj.name}"
        )
        
        node_assembly = nodes.create_assembly_node_tree(
            name = obj.name, 
            iter_list = obj['chain_id_unique'], 
            data_object = data_object
            )
        group = mol.modifiers['MolecularNodes'].node_group
        node = nodes.add_custom(group, node_assembly.name)
        nodes.insert_last_node(group, node)
        
    
    
    return mol


def get_chain_entity_id(file):
    entities = file['entityList']
    chain_names = file['chainNameList']    
    ent_dic = {}
    for i, ent in enumerate(entities):
        for chain_idx in ent['chainIndexList']:
            chain_id = chain_names[chain_idx]
            if  chain_id in ent_dic.keys():
                next
            else:
                ent_dic[chain_id] = i
    
    return ent_dic

def set_atom_entity_id(mol, file):
    mol.add_annotation('entity_id', int)
    ent_dic = get_chain_entity_id(file)
    
    entity_ids = np.array([ent_dic[x] for x in mol.chain_id])
    
    # entity_ids = chain_entity_id[chain_ids]
    mol.set_annotation('entity_id', entity_ids)
    return entity_ids

def open_structure_rcsb(pdb_code, cache_dir = None):
    import biotite.structure.io.mmtf as mmtf
    import biotite.database.rcsb as rcsb
    
    
    file = mmtf.MMTFFile.read(rcsb.fetch(pdb_code, "mmtf", target_path = cache_dir))
    
    # returns a numpy array stack, where each array in the stack is a model in the 
    # the file. The stack will be of length = 1 if there is only one model in the file
    mol = mmtf.get_structure(file, extra_fields = ["b_factor", "charge"], include_bonds = True) 
    set_atom_entity_id(mol, file)
    return mol, file

# operator that calls the function to import the structure from the PDB
class MN_OT_Import_Protein_RCSB(bpy.types.Operator):
    bl_idname = "mn.import_protein_rcsb"
    bl_label = "import_protein_fetch_pdb"
    bl_description = "Download and open a structure from the Protein Data Bank"
    bl_options = {"REGISTER", "UNDO"}

    @classmethod
    def poll(cls, context):
        return not False

    def execute(self, context):
        scene = context.scene
        pdb_code = scene.MN_pdb_code
        cache_dir = scene.MN_cache_dir
        if not scene.MN_cache:
            cache_dir = None
        
        mol = load(
            pdb_code=pdb_code,
            centre=scene.MN_import_centre, 
            del_solvent=scene.MN_import_del_solvent,
            style=scene.MN_import_style,
            cache_dir=cache_dir, 
            build_assembly = scene.MN_import_build_assembly
        )
        
        bpy.context.view_layer.objects.active = mol
        self.report({'INFO'}, message=f"Imported '{pdb_code}' as {mol.name}")
        
        return {"FINISHED"}

def panel(layout, scene):
    
    layout.label(text = "Download from PDB", icon="IMPORT")
    layout.separator()
    row_import = layout.row()
    row_import.prop(scene, 'MN_pdb_code')
    row_import.operator('mn.import_protein_rcsb', text='Download')
    layout.separator()
    layout.label(text = "Options", icon = "MODIFIER")
    options = layout.column(align = True)
    options.prop(scene, "MN_import_style")
    options.separator()
    row = options.row()
    row.prop(scene, 'MN_cache', text="")
    col = row.column()
    col.prop(scene, 'MN_cache_dir', text = "Cache")
    col.enabled = scene.MN_cache
    grid = options.grid_flow()
    for property in ['MN_import_build_assembly', 'MN_import_centre', 'MN_import_del_solvent']:
        grid.prop(scene, property)
    


import bpy
import warnings
from .. import assembly
from .load import create_molecule
from ..blender import (
    nodes
)

bpy.types.Scene.MN_import_local_path = bpy.props.StringProperty(
    name = 'File', 
    description = 'File path of the structure to open', 
    options = {'TEXTEDIT_UPDATE'}, 
    subtype = 'FILE_PATH', 
    maxlen = 0
    )
bpy.types.Scene.MN_import_local_name = bpy.props.StringProperty(
    name = 'Name', 
    description = 'Name of the molecule on import', 
    options = {'TEXTEDIT_UPDATE'}, 
    default = 'NewMolecule', 
    maxlen = 0
    )


def load(
    file_path,                    
    name = "Name",                      
    centre = False,                    
    del_solvent = True,                    
    style = 'spheres',                    
    setup_nodes = True
    ): 
    from biotite import InvalidFileError
    import biotite.structure as struc
    import os
    
    file_path = os.path.abspath(file_path)
    file_ext = os.path.splitext(file_path)[1]
    
    if file_ext == '.pdb':
        mol, file = open_structure_local_pdb(file_path)
        try:
            transforms = assembly.pdb.PDBAssemblyParser(file).get_assemblies()
        except InvalidFileError:
            transforms = None

    elif file_ext == '.pdbx' or file_ext == '.cif':
        mol, file = open_structure_local_pdbx(file_path)
        try:
            transforms = assembly.cif.CIFAssemblyParser(file).get_assemblies()
        except InvalidFileError:
            transforms = None
        
    else:
        warnings.warn("Unable to open local file. Format not supported.")
    # if bonds chosen but no bonds currently exist (mn.bonds is None)
    # then attempt to find bonds by distance
    if not mol.bonds:
        mol.bonds = struc.connect_via_distances(mol[0], inter_residue=True)
    
    if not (file_ext == '.pdb' and file.get_model_count() > 1):
        file = None
        
    
    mol, coll_frames = create_molecule(
        array = mol,
        name = name,
        file = file,
        calculate_ss = True,
        centre = centre,
        del_solvent = del_solvent
        )
    
    # setup the required initial node tree on the object 
    if setup_nodes:
        nodes.create_starting_node_tree(
            object = mol,
            coll_frames = coll_frames,
            style = style
            )
    
    if transforms:
        mol['biological_assemblies'] = transforms
        
    return mol


def open_structure_local_pdb(file_path):
    import biotite.structure.io.pdb as pdb
    
    file = pdb.PDBFile.read(file_path)
    
    # returns a numpy array stack, where each array in the stack is a model in the 
    # the file. The stack will be of length = 1 if there is only one model in the file
    mol = pdb.get_structure(file, extra_fields = ['b_factor', 'charge'], include_bonds = True)
    return mol, file

def open_structure_local_pdbx(file_path):
    import biotite.structure as struc
    import biotite.structure.io.pdbx as pdbx
    from biotite import InvalidFileError
    
    file = pdbx.PDBxFile.read(file_path)
    
    # returns a numpy array stack, where each array in the stack is a model in the 
    # the file. The stack will be of length = 1 if there is only one model in the file
    
    # Try to get the structure, if no structure exists try to get a small molecule
    try:
        mol  = pdbx.get_structure(file, extra_fields = ['b_factor', 'charge'])
    except InvalidFileError:
        mol = pdbx.get_component(file)

    # pdbx doesn't include bond information apparently, so manually create them here
    if not mol.bonds:
        mol[0].bonds = struc.bonds.connect_via_residue_names(mol[0], inter_residue = True)
    return mol, file

# operator that calls the function to import the structure from a local file
class MN_OT_Import_Protein_Local(bpy.types.Operator):
    bl_idname = "mn.import_protein_local"
    bl_label = "Load"
    bl_description = "Open a local structure file"
    bl_options = {"REGISTER", "UNDO"}

    @classmethod
    def poll(cls, context):
        return not False

    def execute(self, context):
        scene = context.scene
        file_path = scene.MN_import_local_path
        
        mol = load(
            file_path=file_path, 
            name=scene.MN_import_local_name, 
            centre=scene.MN_import_centre, 
            del_solvent=scene.MN_import_del_solvent, 
            style=scene.MN_import_style, 
            setup_nodes=True
            
            )
        
        # return the good news!
        bpy.context.view_layer.objects.active = mol
        self.report({'INFO'}, message=f"Imported '{file_path}' as {mol.name}")
        return {"FINISHED"}

    def invoke(self, context, event):
        return self.execute(context)

def panel(layout, scene):
    layout.label(text = "Load a Local File", icon='FILE_TICK')
    layout.separator()
    row_name = layout.row(align = False)
    row_name.prop(scene, 'MN_import_local_name')
    row_name.operator('mn.import_protein_local')
    row_import = layout.row()
    row_import.prop(scene, 'MN_import_local_path')
    layout.separator()
    layout.label(text = "Options", icon = "MODIFIER")
    layout.prop(scene, "MN_import_style")
    grid = layout.grid_flow()
    
    grid.prop(scene, 'MN_import_build_assembly')
    grid.prop(scene, 'MN_import_centre', icon_value=0)
    grid.prop(scene, 'MN_import_del_solvent', icon_value=0)


import bpy
import numpy as np
import os
from ..blender import nodes

bpy.types.Scene.MN_import_density_nodes = bpy.props.BoolProperty(
    name = "Setup Nodes", 
    description = "Creating starting node tree for imported map.",
    default = True
    )
bpy.types.Scene.MN_import_density_invert = bpy.props.BoolProperty(
    name = "Invert Data", 
    description = "Invert the values in the map. Low becomes high, high becomes low.",
    default = False
    )
bpy.types.Scene.MN_import_density = bpy.props.StringProperty(
    name = 'File', 
    description = 'File path for the map file.', 
    subtype = 'FILE_PATH', 
    maxlen = 0
    )
bpy.types.Scene.MN_import_density_name = bpy.props.StringProperty(
    name = 'Name', 
    description = 'Name for the new density object.', 
    default = 'NewDensityObject', 
    maxlen = 0
    )

bpy.types.Scene.MN_import_density_style = bpy.props.EnumProperty(
    name = 'Style', 
    items = (
        ('density_surface', 'Surface', 'A mesh surface based on the specified threshold', 0),
        ('density_wire', 'Wire', 'A wire mesh surface based on the specified threshold', 1)
    )
)

def map_to_grid(file: str, invert: bool = False):
    """
    Reads an MRC file and converts it into a pyopenvdb FloatGrid object.

    This function reads a file in MRC format, and converts it into a pyopenvdb FloatGrid object,
    which can be used to represent volumetric data in Blender.

    Parameters
    ----------
    file : str
        The path to the MRC file.
    invert : bool, optional
        Whether to invert the data from the grid, defaulting to False. Some file types
        such as EM tomograms have inverted values, where a high value == low density.

    Returns
    -------
    pyopenvdb.FloatGrid
        A pyopenvdb FloatGrid object containing the density data.
    """
    import mrcfile
    import pyopenvdb as vdb

    volume = mrcfile.read(file)
    
    dataType = volume.dtype
    
    # enables different grid types
    
    if dataType == "float32" or dataType == "float64":
        grid = vdb.FloatGrid()
    elif dataType == "int8" or dataType == "int16" or dataType == "int32":
        volume = volume.astype('int32')
        grid = vdb.Int32Grid()
    elif dataType == "int64":
        grid = vdb.Int64Grid()

    if invert:
        volume = np.max(volume) - volume
    
    try:
        grid.copyFromArray(volume)
    except ValueError:
        print(f"Grid data type '{volume.dtype}' is an unsupported type.")
    
    grid.gridClass = vdb.GridClass.FOG_VOLUME
    grid.name = 'density'
    return grid


def path_to_vdb(file: str):
    """
    Convert a file path to a corresponding VDB file path.

    Parameters
    ----------
    file : str
        The path of the original file.

    Returns
    -------
    str
        The path of the corresponding VDB file.
    """
    # Set up file paths
    folder_path = os.path.dirname(file)
    name = os.path.basename(file).split(".")[0]
    file_name = name + '.vdb'
    file_path = os.path.join(folder_path, file_name)
    return file_path

    

def map_to_vdb(
    file: str, 
    invert: bool = False, 
    world_scale=0.01, 
    overwrite=False
    ) -> str:
    """
    Converts an MRC file to a .vdb file using pyopenvdb.

    Parameters
    ----------
    file : str
        The path to the input MRC file.
    invert : bool, optional
        Whether to invert the data from the grid, defaulting to False. Some file types
        such as EM tomograms have inverted values, where a high value == low density.
    world_scale : float, optional
        The scaling factor to apply to the voxel size of the input file. Defaults to 0.01.
    overwrite : bool, optional
        If True, the .vdb file will be overwritten if it already exists. Defaults to False.

    Returns
    -------
    str
        The path to the converted .vdb file.
    """
    import mrcfile
    import pyopenvdb as vdb

    file_path = path_to_vdb(file)
    
    # If the map has already been converted to a .vdb and overwrite is False, return that instead
    if os.path.exists(file_path) and not overwrite:
        return file_path

    # Read in the MRC file and convert it to a pyopenvdb grid
    grid = map_to_grid(file, invert=invert)
    
    # Read the voxel size from the MRC file and convert it to a numpy array
    with mrcfile.open(file) as mrc:
        voxel_size = np.array([mrc.voxel_size.x, mrc.voxel_size.y, mrc.voxel_size.z])
    
    # Rotate and scale the grid for import into Blender
    grid.transform.rotate(np.pi / 2, vdb.Axis(1))
    grid.transform.scale(np.array((-1, 1, 1)) * world_scale * voxel_size)
    
    # Write the grid to a .vdb file
    vdb.write(file_path, grid)
    
    # Return the path to the output file
    return file_path


def vdb_to_volume(file: str) -> bpy.types.Object:
    """
    Imports a VDB file as a Blender volume object.

    Parameters
    ----------
    file : str
        Path to the VDB file.

    Returns
    -------
    bpy.types.Object
        A Blender object containing the imported volume data.
    """
    # extract name of file for object name
    name = os.path.basename(file).split('.')[0]
    
    # import the volume object
    bpy.ops.object.volume_import(
        filepath=file, 
        files=[], 
        scale=[1, 1, 1], 
        rotation=[0, 0, 0]
    )
    
    # get reference to imported object and return
    vol = bpy.context.scene.objects[name]
    return vol



def load(
    file: str, 
    name: str = None, 
    style = 'surface', 
    setup_nodes = True, 
    invert: bool = False, 
    world_scale: float = 0.01
    ) -> bpy.types.Object:
    """
    Loads an MRC file into Blender as a volumetric object.

    Parameters
    ----------
    file : str
        Path to the MRC file.
    name : str, optional
        If not None, renames the object with the new name.
    invert : bool, optional
        Whether to invert the data from the grid, defaulting to False. Some file types
        such as EM tomograms have inverted values, where a high value == low density.
    world_scale : float, optional
        Scale of the object in the world. Defaults to 0.01.

    Returns
    -------
    bpy.types.Object
        The loaded volumetric object.
    """
    # Convert MRC file to VDB format
    vdb_file = map_to_vdb(file, invert=invert, world_scale=world_scale)
    
    # Import VDB file into Blender
    vol_object = vdb_to_volume(vdb_file)
    vol_object.mn['molecule_type'] = 'density'
    
    if name:
        # Rename object to specified name
        vol_object.name = name
    
    if setup_nodes:
        nodes.create_starting_nodes_density(vol_object, style = style)
    
    return vol_object


class MN_OT_Import_Map(bpy.types.Operator):
    bl_idname = "mn.import_density"
    bl_label = "Load"
    bl_description = "Import a EM density map into Blender"
    bl_options = {"REGISTER"}

    @classmethod
    def poll(cls, context):
        return True

    def execute(self, context):
        scene = context.scene
        load(
            file = scene.MN_import_density, 
            invert = scene.MN_import_density_invert, 
            setup_nodes=scene.MN_import_density_nodes, 
            style = scene.MN_import_density_style
            )
        return {"FINISHED"}

def panel(layout, scene):
    layout.label(text = 'Load EM Map', icon='FILE_TICK')
    layout.separator()
    
    row = layout.row()
    row.prop(scene, 'MN_import_density_name')
    row.operator('mn.import_density')
    
    layout.prop(scene, 'MN_import_density')
    layout.separator()
    col = layout.column()
    col.alignment = "LEFT"
    col.scale_y = 0.5
    label = f"\
    An intermediate file will be created: {path_to_vdb(scene.MN_import_density)}.\
    Please do not delete this file or the volume will not render.\
    Move the original .map file to change this location.\
    "
    for line in label.strip().split('    '):
        col.label(text=line)
    
    layout.separator()
    layout.label(text = "Options", icon = "MODIFIER")
    layout.prop(scene, "MN_import_density_style")
    
    grid = layout.grid_flow()
    grid.prop(scene, 'MN_import_density_nodes')
    grid.prop(scene, 'MN_import_density_invert')


import numpy as np
from pathlib import Path
import bpy
from .. import assembly
from ..blender import (
    obj, nodes, coll
)
from ..io.load import create_molecule
from . import bcif
from ..color import random_rgb

bpy.types.Scene.mol_import_cell_pack_path = bpy.props.StringProperty(
    name = 'File', 
    description = 'File to import (.cif, .bcif)', 
    subtype = 'FILE_PATH', 
    maxlen = 0
    )
bpy.types.Scene.mol_import_cell_pack_name = bpy.props.StringProperty(
    name = 'Name', 
    description = 'Name of the created object.', 
    default = 'NewCellPackModel', 
    maxlen = 0
    )


def load(
    file_path, 
    name='NewCellPackModel', 
    node_tree=True, 
    world_scale=0.01, 
    fraction: float = 1, 
    instance_nodes=True
    ):
    ensemble, chain_collection = parse(file_path, name=name, instance_nodes=instance_nodes)
    starting_node_tree(ensemble, chain_collection, name=name, fraction=fraction)
    return ensemble


def parse(file, name="NewModel", get_transforms=True, instance_nodes=True):
    import biotite.structure.io.pdbx as pdbx

    print("openfile",file)
    if Path(file).suffix in (".bcif", ".bin"):
        mol, transforms = bcif.parse(file)
        # get transforms and create data / CellPack Object
        if get_transforms:
            obj_data = assembly.mesh.create_data_object(transforms, name=name, collection=coll.mn())
    else:
        file_open = pdbx.PDBxFile.read(file)
        print("file_open ok")
        mol = pdbx.get_structure(file_open,  model=1, extra_fields=['label_entity_id'])
        print("loaded mol", len(mol))
        transforms = assembly.cif.CIFAssemblyParser(file_open).get_assemblies()
        print("loaded transforms", len(transforms))
        
        # get transforms and create data / CellPack Object
        transforms_array = assembly.mesh.get_transforms_from_dict(transforms)
        obj_data = assembly.mesh.create_data_object(transforms_array, name=name, collection=coll.mn())
    
    chain_names = np.unique(mol.chain_id)
    
    obj_data['chain_id'] = chain_names

    coll_cellpack = coll.cellpack(f"{name}")

    for i, chain in enumerate(chain_names):
        atoms = mol[mol.chain_id == chain]
        print(chain, atoms.res_name[0])
        # if atoms.res_name[0] == 'LIP': 
        #     continue
        mol_object, coll_frames = create_molecule(
            array=atoms,
            name=f"{str(i).rjust(4, '0')}_{chain}",
            collection=coll_cellpack
            )

        colors = np.tile(random_rgb(i), (len(atoms), 1))

        obj.add_attribute(mol_object, name="Color", data=colors, type="FLOAT_COLOR", overwrite=True)
        if instance_nodes:
            nodes.create_starting_node_tree(mol_object, name = f"MN_pack_instance_{name}", set_color=False)

    return obj_data, coll_cellpack


def starting_node_tree(ensemble, coll_cellpack, name = "CellPackModel", fraction: float = 1.0, fallback=False):
    # ensure there is a geometry nodes modifier called 'MolecularNodes' that is created and applied to the object
    node_mod = ensemble.modifiers.get('MolecularNodes')
    if not node_mod:
        node_mod = ensemble.modifiers.new("MolecularNodes", "NODES")

    ensemble.modifiers.active = node_mod
    
    group = nodes.new_group(name = f"MN_cellpack_{name}", fallback=False)
    node_mod.node_group = group
    
    node_pack = nodes.add_custom(group, 'MN_pack_instances', location=[-100,0])
    node_pack.inputs['Collection'].default_value = coll_cellpack
    node_pack.inputs['Fraction'].default_value = fraction
    
    link = group.links.new
    link(nodes.get_input(group).outputs[0], node_pack.inputs[0])
    link(node_pack.outputs[0], nodes.get_output(group).inputs[0])


class MN_OT_Import_Cell_Pack(bpy.types.Operator):
    bl_idname = "mol.import_cell_pack"
    bl_label = "Load"
    bl_description = ""
    bl_options = {"REGISTER"}

    @classmethod
    def poll(cls, context):
        return True

    def execute(self, context):
        s = context.scene
        load(
            file_path = s.mol_import_cell_pack_path, 
            name = s.mol_import_cell_pack_name, 
            node_tree = True, 
            fraction = s.mol_import_cell_pack_fraction
        )
        return {"FINISHED"}

def panel(layout, scene):
    layout.label(text = "Load CellPack Model", icon='FILE_TICK')
    layout.separator()
    row_import = layout.row()
    row_import.prop(scene, 'mol_import_cell_pack_name')
    layout.prop(scene, 'mol_import_cell_pack_path')
    row_import.operator('mol.import_cell_pack')



# Import necessary modules or classes
# Example:
# from .module1 import Class1
# from .module2 import Class2

# Define any variables or constants
# Example:
# VERSION = '1.0'

# Define any functions or classes that should be accessible from outside the submodule
# Example:
# def my_function():
#     pass

# Optionally, you can use the __all__ variable to specify the public interface of the submodule
# Example:
# __all__ = ['Class1', 'my_function']

__all__ = ['pdb', 'star', 'pack', 'md', 'density']



import bpy
import numpy as np

def create_object(
    locations: np.ndarray, 
    edges = [], 
    faces = [], 
    name: str = 'NewObject', 
    collection: bpy.types.Collection = None
    ) -> bpy.types.Object:
    """
    Create a mesh with the given name in the given collection, using the supplied
    vertex locations and, if provided, edges.

    Parameters
    ----------
    name : str
        The name of the mesh object to be created.
    collection : bpy.types.Collection
        The collection to which the mesh object will be added.
    locations : array-like
        The list of vertex locations for the mesh, an nx3 np array of locations.
        Each element in the list represents a 3D point (x, y, z) for a vertex.
    bonds : list of tuples, optional
        The list of vertex index pairs representing bonds as edges for the mesh.
        Each tuple should contain two vertex indices (e.g., (index1, index2)).
    faces : list of tupls, options
        List of vertex groups, each group will become a distrinct face.

    Returns
    -------
    bpy.types.Object
        The newly created mesh object.

    Notes
    -----
    - The 'name' should be a unique identifier for the created mesh object.
    - The 'locations' list should contain at least three 3D points to define a 3D triangle.
    - If 'edges' are not provided, the mesh will have no edges.
    - If 'edges' are provided, they should be valid vertex indices within the 'locations' list.

    Example
    -------
    ```python
    # Create a mesh object named "MyMesh" in the collection "MyCollection"
    # with vertex locations and bond edges.
    locations = [[0.0, 0.0, 0.0], [1.0, 0.0, 0.0], [1.0, 1.0, 0.0]]
    edges = [(0, 1), (1, 2), (2, 0)]
    my_object = create_object(lcoations, edges, name='MySimpleObject')
    ```
    """
    # Have to first create a mesh, then fill th emesh with data from verts, edges and faces
    mesh = bpy.data.meshes.new(name)
    mesh.from_pydata(locations, edges, faces=[])
    
    # create a new object that has a name and links to the created mesh
    object = bpy.data.objects.new(name, mesh)
    if not collection:
        collection = bpy.data.collections['Collection']
    collection.objects.link(object)
    
    # assign various simple attributes / properties to the object
    object['type'] = 'molecule'
    return object


def add_attribute(object: bpy.types.Object, name: str, data: np.ndarray, type="FLOAT", domain="POINT", overwrite: bool = False):
    """
    Add an attribute to the given object's geometry on the given domain.

    Parameters
    ----------
        object : bpy.types.Object
            The object to which the attribute will be added.
        name : str
            The name of the attribute.
        data : array-like
            The data to be assigned to the attribute. "FLOAT_VECTOR" and "FLOAT_COLOR" entries should be of length 3 and 4 respectively. 
        type : str, optional, default: "FLOAT"
            The data type of the attribute. Possible values are "FLOAT", "FLOAT_VECTOR", "FLOAT_COLOR", "INT", or "BOOLEAN".
        domain : str, optional, default: "POINT"
            The domain to which the attribute is added. Possible values are "POINT" or other domains supported
            by the object.

    Returns
    -------
        Any
            The newly created attribute, which can be further manipulated or used in the 3D environment.

    Notes
    -----
        - The function supports adding both scalar and vector attributes.
        - The "FLOAT_VECTOR" attribute requires the input data to be a 1D array, and it will be reshaped internally
          to represent vectors with 3 components (x, y, z).
    """
    att = object.data.attributes.get(name)
    if not att or not overwrite:
        att = object.data.attributes.new(name, type, domain)
    
    types = {
        'FLOAT_VECTOR': 'vector',
        'FLOAT_COLOR': 'color', 
        'INT': 'value', 
        'FLOAT': 'value',
        'BOOLEAN': 'value' 
        
    }
    
    att.data.foreach_set(types[type], data.reshape(-1).copy(order='c'))
    
    return att

def get_attribute(obj: bpy.types.Object, att_name='position') -> np.array:
    """
    Retrieve an attribute from the object as a NumPy array.

    Parameters
    ----------
    obj : bpy.types.Object
        The Blender object from which the attribute will be retrieved.
    att_name : str, optional
        The name of the attribute to retrieve. Default is 'position'.
        
    Returns
    -------
    np.array
        The attribute data as a NumPy array.
    
    Notes
    -----
    - This function retrieves the specified attribute from the object and returns it as a NumPy array.
    - The function assumes that the attribute data type is one of ['INT', 'FLOAT', 'BOOLEAN', 'FLOAT_VECTOR'].

    Example
    -------
    ```python
    # Assuming 'my_object' is a Blender object with an attribute named 'my_attribute'
    attribute_data = get_attribute(my_object, 'my_attribute')
    ```
    """

    # Get the attribute from the object's mesh
    att = obj.to_mesh().attributes[att_name]

    # Map attribute values to a NumPy array based on the attribute data type
    if att.data_type in ['INT', 'FLOAT', 'BOOLEAN']:
        # Define the mapping of Blender data types to NumPy data types
        d_type = {'INT': int, 'FLOAT': float, 'BOOLEAN': bool}
        # Convert attribute values to a NumPy array with the appropriate data type
        att_array = np.array(list(map(lambda x: x.value, att.data.values())), dtype=d_type.get(att.data_type))
    elif att.data_type == "FLOAT_VECTOR":
        # Convert attribute vectors to a NumPy array
        att_array = np.array(list(map(lambda x: x.vector, att.data.values())))
    elif att.data_type == "FLOAT_COLOR":
        att_array = np.array(list(map(lambda x: x.color, att.data.values())))
    else:
        # Unsupported data type, return an empty NumPy array
        att_array = np.array([])

    return att_array


def set_position(object, locations: np.ndarray):
    """
    Update the vertex positions of a Blender object.

    Parameters
    ----------
    object : bpy.types.Object
        The Blender object whose vertex positions need to be updated.
    locations : numpy.ndarray, optional
        An array containing the new vertex positions. Default is an empty array.

    Returns
    -------
    None

    Raises
    ------
    TypeError
        If `object` is not of type `bpy.types.Object`.
        If `locations` is not of type `numpy.ndarray`.
    ValueError
        If the shape of `locations` is not (n, 3), where n is the number of vertices.
    AttributeError
        If the object's data block does not have a 'position' attribute.

    Notes
    -----
    The `locations` array should be of shape (n, 3), where n is the number of vertices.
    The `object` should have a data block containing a 'position' attribute.

    Example
    -------
    set_position(obj, np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]]))
    """
    # Check if the input object is valid
    if not isinstance(object, bpy.types.Object):
        raise TypeError("Expected 'object' to be a bpy.types.Object")

    # Check if the input locations array is valid
    if not isinstance(locations, np.ndarray):
        raise TypeError("Expected 'locations' to be a numpy.ndarray")

    if locations.shape[1] != 3:
        raise ValueError("The 'locations' array should be of shape (n, 3)")

    # Check if the object has a 'position' attribute
    if 'position' not in object.data.attributes:
        raise AttributeError("The object's data block must have a 'position' attribute")

    pos = object.data.attributes['position']

    # Ensure the locations array is flattened and compatible with the 'vector' attribute
    pos.data.foreach_set('vector', locations.reshape(-1))

    # Update the object's data
    object.data.update()



import numpy as np
import bpy
from .. import color
from ..blender import (
    obj, coll, nodes
)

bpy.types.Scene.MN_import_oxdna_topology = bpy.props.StringProperty(
    name = 'Toplogy', 
    description = 'File path for the topology to import (.top)', 
    subtype = 'FILE_PATH', 
    maxlen = 0
    )
bpy.types.Scene.MN_import_oxdna_trajectory = bpy.props.StringProperty(
    name = 'Trajectory', 
    description = 'File path for the trajectory to import (.oxdna / .dat)', 
    subtype = 'FILE_PATH', 
    maxlen = 0
    )
bpy.types.Scene.MN_import_oxdna_name = bpy.props.StringProperty(
    name = 'Name', 
    description = 'Name of the created object.', 
    default = 'NewOrigami', 
    maxlen = 0
    )

def base_to_int(bases: np.array) -> np.array:
    """
    Convert an array of DNA bases to their corresponding MN integer values.

    Parameters
    ----------
    bases : np.array
        Array of DNA bases.

    Returns
    -------
    np.array
        Array of corresponding integer values for the DNA bases.
    """
    # Values for internal Molecular Nodes use. Defined in data.py
    base_lookup = {
        'A': 30, 
        'C': 31,
        'G': 32, 
        'T': 33
    }
    
    ints = np.array([base_lookup.get(base, -1) for base in bases])
    
    return ints

def is_new_topology(filepath):
    with open(filepath) as f:
        firstline = f.readline()
    
    return "5 -> 3" in firstline

def read_topology_new(filepath):
    with open(filepath, 'r') as file:
        contents = file.read()
    
    lines = np.array(contents.split('\n'))
    
    def read_seq_line(line):
        sequence = line.split(" ")[0]
        return np.array([c for c in sequence])
    
    strands = []
    counter = 0
    
    for i, line in enumerate(lines[1:]):
        bases = read_seq_line(line)
        arr = np.zeros((len(bases), 4), dtype=int)
        idx = np.array(range(len(bases)), dtype = int)
        arr[:, 0] = i + 1 # strand ID
        arr[:, 1] = base_to_int(bases) # base
        bond_3 = idx -1 + counter
        bond_5 = idx + 1 + counter
        bond_3[0] = -1
        bond_5[-1] = -1
        arr[:, 2] = bond_3
        arr[:, 3] = bond_5
        
        strands.append(arr)
        counter += len(bases)
    
    return np.vstack(strands)

def read_topology_old(filepath):
    """
    Read the topology from a file and convert it to a numpy array.
    
    
    Strand assignment
    |  Base assignment
    |  |  3' Bonded base to the current base (index based on row)
    |  |  |   5' Bonded base to the current base (index based on row)
    |  |  |   |
    S  B  3'  5'
    S  B  3'  5'
    S  B  3'  5'

    Parameters
    ----------
    filepath : str
        The path to the file containing the topology.

    Returns
    -------
    numpy.ndarray
        The topology as a integer numpy array. Base assignment is (30, 31, 32, 33) where 
        this corresponds to (A, C, G, T) for use inside of Molecular Nodes.

    """
    
    with open(filepath, 'r') as file:
        contents = file.read()

    lines = np.array(contents.split('\n'))
    # metadata = lines[0]
    
    # read the topology from the file sans the first metadata line
    # have to initially read as strings, then convert bases to numeric later
    array_str = np.loadtxt(lines[1:], dtype=str)
    
    # convert the columns to numeric
    array_int = np.zeros(array_str.shape, dtype=int)
    array_int[:, (0, 2, 3)] = array_str[:, (0, 2, 3)].astype(int) # easy convert numeric columns to int
    array_int[:, 1] = base_to_int(array_str[:, 1]) # convert bases (A, C, G, T) to (30, 31, 32, 33)
    
    return array_int

def read_trajectory(filepath):
    """
    Read an oxDNA trajectory file and return an array of frames.
    
    Each frame becomes a 2D array in a stack. Each frame has 5 three-component vectors. 
    The vectors are: (position, base_vector, base_normal, veclocity, angular_velocity), 
    which totals 15 columns in the array. The (velocity, angular_velocity) are optional
    and can sometimes not appear in the trajectory.

    Parameters
    ----------
    filepath : str
        The path to the trajectory file.

    Returns
    -------
    frames : ndarray
        An array of frames, where each frame is a 2D array of positions 

    """
    # Open the file and read its contents
    with open(filepath, 'r') as file:
        contents = file.read()

    # Split the contents into lines
    lines = np.array(contents.split('\n'))
    is_meta = np.char.find(lines, '=') > 0

    group_id = np.cumsum(np.append([True], np.diff(is_meta)))
    groups = np.unique(group_id)
    
    frames = []

    for group in groups:
        mask = group == group_id
        if "=" in lines[mask][0]:
            continue
        
        arr = np.loadtxt(lines[mask])
        frames.append(arr)
    
    return np.stack(frames)

def add_attributes_to_dna_mol(mol, frame, scale_dna = 0.1):
    attributes = ('base_vector', 'base_normal', 'velocity', 'angular_velocity')
    for i, att in enumerate(attributes):
        col_idx = np.array([3, 4, 5]) + i * 3
        
        try:
            data = frame[:, col_idx]
        except IndexError as e:
            print(f"Unable to get {att} attribute from coordinates. Error: {e}")
            continue
        
        if att != "angular_velocity":
            data *= scale_dna
        
        obj.add_attribute(mol, att, data, type="FLOAT_VECTOR")

def toplogy_to_bond_idx_pairs(topology: np.ndarray):
    """
    Convert the given topology array into pairs of indices representing each distinct bond.

    Strand assignment
    |  Base assignment
    |  |  3' Bonded base to the current base (index based on row)
    |  |  |  5' Bonded base to the current base (index based on row)
    |  |  |  |
    1  A -1  1
    1  G  0  2
    1  C  1 -1
    
    The topology above becomes:
    np.array([[0, 1], [2, 1]])
    
    The order of the bond indices doesn't matter to Blender.
    
    Parameters:
    topology (np.ndarray): Numeric numpy array representing the topology.

    Returns:
    np.ndarray: Array of pairs of indices representing each distinct bond.
    """


    # to get pairs of indices which represent each distinct bond, which are needed for 
    # edge creation in Blender, take each bonded column and create a 'bond' with itself
    idx = np.array(list(range(topology.shape[0])))
    bond_3 = np.vstack((idx, topology[:, 2])).reshape((len(idx), 2))
    bond_5 = np.vstack((idx, topology[:, 3])).reshape((len(idx), 2))
    bonds = np.vstack((bond_3, bond_5))

    # drop where either bond is -1 (not bonded) from the bond indices
    mask = bonds == -1
    mask = np.logical_not(mask.any(axis=1)) 
    
    bond_idxs = np.unique(bonds[mask, :], axis = 0)
    

    return np.sort(bond_idxs, axis = 1)

def load(top, traj, name = 'oxDNA', setup_nodes=True, world_scale = 0.01):
    
    # the scale of the oxDNA files seems to be based on nanometres rather than angstrongs 
    # like most structural biology files, so currently adjusting the world_scale to 
    # compensate
    scale_dna = world_scale * 10
    
    # read in the topology and trajectory files
    is_new_top = is_new_topology(top)
    if is_new_top:
        topology = read_topology_new(top)
    else:
        topology = read_topology_old(top)
    
    trajectory = read_trajectory(traj)
    n_frames = trajectory.shape[0]
    
    # creat toplogy object with positions of the first frame, and the bonds from the 
    # topology object
    mol = obj.create_object(
        name=name,
        collection=coll.mn(),
        locations=trajectory[0][:, 0:3] * scale_dna,
        edges=toplogy_to_bond_idx_pairs(topology)
    )
    
    # adding additional toplogy information from the topology and frames objects
    obj.add_attribute(mol, 'res_name', topology[:, 1], "INT")
    obj.add_attribute(mol, 'chain_id', topology[:, 0], "INT")
    obj.add_attribute(mol, 'Color', data=color.color_chains_equidistant(topology[:, 0]), type = 'FLOAT_COLOR')
    add_attributes_to_dna_mol(mol, trajectory[0], scale_dna=scale_dna)
    
    
    # if the 'frames' file only contained one timepoint, return the object without creating
    # any kind of collection for storing multiple frames from a trajectory, and a None 
    # object in place of the frames collection
    if n_frames == 1:
        if setup_nodes:
            nodes.create_starting_node_tree(mol, style="oxdna", set_color=False)
        return mol, None
    
    # create a collection to store all of the frame objects that are part of the trajectory
    # they will contain all of the possible attributes which can be interpolated betewen 
    # frames such as position, base_vector, base_normal, velocity, angular_velocity
    collection = coll.frames(f"{name}_frames", parent=coll.data())
    for i, frame in enumerate(trajectory):
        fill_n = int(np.ceil(np.log10(n_frames)))
        frame_name = f"{name}_frame_{str(i).zfill(fill_n)}"
        frame_mol = obj.create_object(frame[:, 0:3] * scale_dna, name=frame_name, collection=collection)
        add_attributes_to_dna_mol(frame_mol, frame, scale_dna)
    
    if setup_nodes:
        nodes.create_starting_node_tree(mol, coll_frames=collection, style="oxdna", set_color=False)
    
    return mol, collection


class MN_OT_Import_OxDNA_Trajectory(bpy.types.Operator):
    bl_idname = "mn.import_oxdna"
    bl_label = "Load"
    bl_description = "Will import the given file and toplogy."
    bl_options = {"REGISTER"}

    def execute(self, context):
        s = context.scene
        load(
            top = s.MN_import_oxdna_topology,
            traj= s.MN_import_oxdna_trajectory, 
            name= s.MN_import_oxdna_name
        )
        return {"FINISHED"}


def panel(layout, scene):
    layout.label(text = "Load oxDNA File", icon='FILE_TICK')
    layout.separator()
    row = layout.row()
    row.prop(scene, 'MN_import_oxdna_name')
    row.operator('mn.import_oxdna')
    col = layout.column(align=True)
    col.prop(scene, 'MN_import_oxdna_topology')
    col.prop(scene, 'MN_import_oxdna_trajectory')


"""
Importing molecular dynamics trajectories and associated files.
"""

__name__ = "MolecularNodes.trajectory"
__author__ = "Brady Johnston"

import bpy

try:
    import MDAnalysis as mda
except ImportError:
    HAS_mda = False
    import types

    class MockUniverse:
        pass

    mda = types.ModuleType("MDAnalysis")
    mda.Universe = MockUniverse

else:
    HAS_mda = True

from .mda import MDAnalysisSession
from .. import pkg

bpy.types.Scene.MN_import_md_topology = bpy.props.StringProperty(
    name='Topology',
    description='File path for the toplogy file for the trajectory',
    subtype='FILE_PATH',
    maxlen=0
)
bpy.types.Scene.MN_import_md_trajectory = bpy.props.StringProperty(
    name='Trajectory',
    description='File path for the trajectory file for the trajectory',
    subtype='FILE_PATH',
    maxlen=0
)
bpy.types.Scene.MN_import_md_name = bpy.props.StringProperty(
    name='Name',
    description='Name of the molecule on import',
    default='NewTrajectory',
    maxlen=0
)
bpy.types.Scene.MN_import_md_frame_start = bpy.props.IntProperty(
    name="Start",
    description="Frame start for importing MD trajectory",
    default=0
)
bpy.types.Scene.MN_import_md_frame_step = bpy.props.IntProperty(
    name="Step",
    description="Frame step for importing MD trajectory",
    default=1
)
bpy.types.Scene.MN_import_md_frame_stop = bpy.props.IntProperty(
    name="Stop",
    description="Frame stop for importing MD trajectory",
    default=499
)
bpy.types.Scene.MN_md_selection = bpy.props.StringProperty(
    name='Import Filter',
    description='Custom MDAnalysis selection string, removing unselecte atoms. See: "https://docs.mdanalysis.org/stable/documentation_pages/selections.html"',
    default='all'
)
bpy.types.Scene.MN_md_in_memory = bpy.props.BoolProperty(
    name='Memory',
    description='True will load all of the requested frames into the scene and into memory. False will stream the trajectory from a live MDAnalysis session',
    default=False
)
bpy.types.Scene.list_index = bpy.props.IntProperty(
    name="Index for trajectory selection list.",
    default=0
)

def load(
    top, 
    traj, 
    name = "NewTrajectory", 
    style = "spheres",
    selection:str = 'all', 
    start:int = 0, 
    step:int = 1,
    stop:int = 499, 
    subframes:int = 0, 
    custom_selections:dict = {}, 
    in_memory:bool = False
):
    universe = mda.Universe(top, traj)
    
    if in_memory:
        universe.transfer_to_memory(start=start, step=step, stop=stop)
        
    session = MDAnalysisSession()
    
    extra_selections = {}
    for sel in custom_selections:
        extra_selections[sel.name] = sel.selection
    mol = session.show(atoms=universe,
                            name=name,
                            style=style,
                            selection=selection,
                            custom_selections=extra_selections,
                            in_memory=in_memory
                            )
    
    return mol, universe


class MN_OT_Import_Protein_MD(bpy.types.Operator):
    bl_idname = "mn.import_protein_md"
    bl_label = "Import Protein MD"
    bl_description = "Load molecular dynamics trajectory"
    bl_options = {"REGISTER", "UNDO"}

    @classmethod
    def poll(cls, context):
        return True

    def execute(self, context):
        scene = context.scene
        if not pkg.is_current('MDAnalysis'):
            self.report({'ERROR'},
                        message="MDAnalysis is not installed. "
                                "Please install it to use this feature.")
            return {'CANCELLED'}
        top = scene.MN_import_md_topology
        traj = scene.MN_import_md_trajectory
        name = scene.MN_import_md_name

        mol, universe = load(
            top=top,
            traj = traj, 
            name = name,
            style = scene.MN_import_style,
            selection = scene.MN_md_selection, 
            start = scene.MN_import_md_frame_start, 
            stop = scene.MN_import_md_frame_stop, 
            step = scene.MN_import_md_frame_step, 
            custom_selections=scene.trajectory_selection_list,
            in_memory=scene.MN_md_in_memory
            
        )

        bpy.context.view_layer.objects.active = mol

        self.report(
            {'INFO'},
            message=f"Imported '{top}' as {name} "
                    f"with {str(universe.trajectory.n_frames)} "
                    f"frames from '{traj}'."
        )

        return {"FINISHED"}


#### UI

class TrajectorySelectionItem(bpy.types.PropertyGroup):
    """Group of properties for custom selections for MDAnalysis import."""
    bl_idname = "testing"
    
    name: bpy.props.StringProperty(
        name="Attribute Name", 
        description="Attribute", 
        default="custom_selection"
    )
    
    selection: bpy.props.StringProperty(
        name="Selection String", 
        description="String that provides a selection through MDAnalysis", 
        default = "name CA"
    )


# have to manually register this class otherwise the PropertyGroup registration fails
bpy.utils.register_class(TrajectorySelectionItem)
bpy.types.Scene.trajectory_selection_list = bpy.props.CollectionProperty(
    type = TrajectorySelectionItem
)

class MN_UL_TrajectorySelectionListUI(bpy.types.UIList):
    """UI List"""
    
    def draw_item(self, context, layout, data, item, 
                  icon, active_data, active_propname, index):
        custom_icon = "VIS_SEL_11"
        
        if self.layout_type in {'DEFAULT', 'COMPACT'}:
            layout.label(text = item.name, icon = custom_icon)
        
        elif self.layout_type in {'GRID'}:
            layout.alignment = 'CENTER'
            layout.label(text = "", icon = custom_icon)
            

class TrajectorySelection_OT_NewItem(bpy.types.Operator):
    """Add a new custom selection to the list."""
    
    bl_idname = "trajectory_selection_list.new_item"
    bl_label = "+"
    
    def execute(self, context):
        context.scene.trajectory_selection_list.add()
        return {'FINISHED'}

class TrajectorySelection_OT_DeleteIem(bpy.types.Operator):
    
    bl_idname = "trajectory_selection_list.delete_item"
    bl_label = "-"
    
    @classmethod
    def poll(cls, context):
        return context.scene.trajectory_selection_list
    def execute(self, context):
        my_list = context.scene.trajectory_selection_list
        index = context.scene.list_index
        
        my_list.remove(index)
        context.scene.list_index = min(max(0, index - 1), len(my_list) - 1)
        
        return {'FINISHED'}

def custom_selections(layout, scene):
    layout.label(text="Custom Selections")
    row = layout.row(align=True)
    
    row = row.split(factor = 0.9)
    row.template_list('MN_UL_TrajectorySelectionListUI', 'A list', scene, 
                        "trajectory_selection_list", scene, "list_index", rows=3)
    col = row.column()
    col.operator('trajectory_selection_list.new_item', icon="ADD", text="")
    col.operator('trajectory_selection_list.delete_item', icon="REMOVE", text="")
    if scene.list_index >= 0 and scene.trajectory_selection_list:
        item = scene.trajectory_selection_list[scene.list_index]
        
        col = layout.column(align=False)
        col.separator()
        
        col.prop(item, "name")
        col.prop(item, "selection")

def panel(layout, scene):
    layout.label(text = "Load MD Trajectories", icon='FILE_TICK')
    layout.separator()
    col = layout.column(align=True)
    row_import = col.row()
    row_import.prop(scene, 'MN_import_md_name')
    row_import.operator('mn.import_protein_md', text = "Load")
    col.separator()
    col.prop(scene, 'MN_import_md_topology')
    col.prop(scene, 'MN_import_md_trajectory')
    
    layout.separator()
    layout.label(text = "Options", icon = "MODIFIER")
    layout.prop(scene, "MN_import_style")
    layout.prop(scene, 'MN_md_selection')
    row_frame = layout.row(heading = "Frames", align = True)
    row_frame.prop(scene, 'MN_md_in_memory')
    row = row_frame.row(align=True)
    row.prop(scene, 'MN_import_md_frame_start')
    row.prop(scene, 'MN_import_md_frame_step')
    row.prop(scene, 'MN_import_md_frame_stop')
    row.enabled = scene.MN_md_in_memory
    custom_selections(layout, scene)